# -*- coding: utf-8 -*-
# pylint: disable = import-outside-toplevel
"""Tests of ``lib.form`` functions and class methods."""

from unittest import TestCase


# =============================================================================
class ULibFormGetAction(TestCase):
    """Unit test class for :func:`lib.form.get_action`."""

    # -------------------------------------------------------------------------
    def test_it(self):
        """[u:lib.form.get_action]"""
        from sys import version_info
        from pyramid.testing import DummyRequest
        from ..lib.form import get_action

        request = DummyRequest(POST={})
        action, items = get_action(request)
        self.assertEqual(action, '')
        self.assertEqual(len(items), 0)

        request = DummyRequest(POST={'run!1.x': '8'})
        action, items = get_action(request)
        self.assertEqual(action, 'run!1')
        self.assertEqual(len(items), 1)
        self.assertEqual(items[0], '1')

        request = DummyRequest(POST={
            'run!éÔ.x'.encode('utf8'): '8'})
        action, items = get_action(request)
        item = 'run!éÔ' if version_info > (3, 0) else 'run!éÔ'.encode('utf8')
        self.assertEqual(action, item)
        self.assertEqual(len(items), 1)
        self.assertEqual(items[0], item[4:])

        request = DummyRequest(POST={'run!#.x': '8'})
        action = get_action(request)[0]
        self.assertEqual(action, '')
        self.assertTrue(request.session.pop_flash('alert'))

        request = DummyRequest(POST={'run!#.x': '8', '#1': '1', '#2': '1'})
        action, items = get_action(request)
        self.assertEqual(action, 'run!#')
        self.assertEqual(len(items), 2)
        self.assertIn(items[0], ('1', '2'))

        request = DummyRequest(POST={
            'run!#.x': '8', bytes('#éÔ'.encode('utf8')): '1'})
        action, items = get_action(request)
        self.assertEqual(len(items), 1)
        self.assertEqual(
            items[0], version_info > (3, 0) and 'éÔ' or 'éÔ'.encode('utf8'))


# =============================================================================
class ULibFormSameAs(TestCase):
    """Unit test class for :class:`lib.form.SameAs`."""

    # -------------------------------------------------------------------------
    def test_it(self):
        """[u:lib.form.SameAs]"""
        from colander import Invalid
        from pyramid.testing import DummyRequest
        from ..lib.form import SameAs

        same_as = SameAs(DummyRequest(), 'foo')
        self.assertRaises(Invalid, same_as, 'login', 'bar')


# =============================================================================
class ULibFormButton(TestCase):
    """Unit test class for :func:`lib.form.button`."""

    # -------------------------------------------------------------------------
    def test_with_no_label(self):
        """[u:lib.form.button] with no label"""
        from ..lib.form import button
        from webhelpers2.html import literal

        test_button = button(url='url', src='src')
        expected_result = \
            literal('<a href="url"><img src="src" alt="None"/></a> ')
        self.assertEqual(test_button, expected_result)

    # -------------------------------------------------------------------------
    def test_regular_button(self):
        """[u:lib.form.button] regular button"""
        from ..lib.form import button
        from webhelpers2.html import literal

        test_button = button(
            url='url', label='label', src='src', title='title', class_='value')
        expected_result = \
            literal('<a href="url" title="title" class="value">'
                    '<img src="src" alt="label"/>label</a> ')
        self.assertEqual(test_button, expected_result)


# =============================================================================
class ULibFormGridItem(TestCase):
    """Unit test class for :func:`lib.form.grid_item`."""

    # -------------------------------------------------------------------------
    def test_it(self):
        """[u:lib.form.grid_item]"""
        from ..lib.form import grid_item

        self.assertFalse(grid_item(None, 'label', None))
        self.assertEqual(
            grid_item(
                None, 'label', 'html', True, 'help', 'erreur', clear=True),
            '<div class="cioError">'
            '<label><strong>label<span>*</span></strong></label>'
            '<div>html<em> help</em><strong> erreur</strong></div>'
            '<div class="cioClear"></div></div>')
        self.assertEqual(
            grid_item(
                None, 'label', 'html', False, 'help', 'erreur',
                class_='noClass'),
            '<div class="noClass cioError">'
            '<label><strong>label</strong></label>'
            '<div>html<em> help</em><strong> erreur</strong></div></div>')


# =============================================================================
class ULibFormForm(TestCase):
    """Unit test class for :class:`lib.form.Form`."""
    # pylint: disable = too-many-public-methods

    # -------------------------------------------------------------------------
    def setUp(self):
        """Set up test application."""
        from pyramid.testing import setUp
        from pyramid.csrf import SessionCSRFStoragePolicy

        self.configurator = setUp()
        self.configurator.set_csrf_storage_policy(SessionCSRFStoragePolicy())

    # -------------------------------------------------------------------------
    def tearDown(self):
        """Undo the effects of ``pyramid.testing.setUp()``."""
        from pyramid.testing import tearDown

        tearDown()

    # -------------------------------------------------------------------------
    @classmethod
    def schema(cls):
        """Return a simple Colander schema."""
        from colander import Mapping, SchemaNode, String, Boolean, Length

        schema = SchemaNode(Mapping())
        schema.add(SchemaNode(String(), name='login', validator=Length(min=2)))
        schema.add(SchemaNode(Boolean(), name='remember', missing=False))
        return schema

    # -------------------------------------------------------------------------
    def test_init(self):
        """[u:lib.form.Form.__init__]"""
        from collections import namedtuple
        from pyramid.testing import DummyRequest
        from ..lib.form import Form

        request = DummyRequest()
        form = Form(request)
        self.assertFalse(form.values)

        form = Form(request, self.schema())
        self.assertFalse(form.values)

        form = Form(request, self.schema(), defaults={'login': 'user1'})
        self.assertIn('login', form.values)
        self.assertEqual(form.values['login'], 'user1')

        user = namedtuple('User', 'login')(login='user1')
        form = Form(request, self.schema(), obj=user)
        self.assertIn('login', form.values)
        self.assertEqual(form.values['login'], 'user1')

    # -------------------------------------------------------------------------
    def test_validate_request_no_post(self):
        """[u:lib.form.Form.validate] request with no post"""
        from pyramid.testing import DummyRequest
        from ..lib.form import Form

        form = Form(DummyRequest())
        self.assertFalse(form.validate())

    # -------------------------------------------------------------------------
    def test_validate_ko(self):
        """[u:lib.form.Form.validate] KO"""
        from colander import Mapping, SchemaNode, String, Boolean, Length
        from pyramid.testing import DummyRequest
        from ..lib.form import Form

        schema = SchemaNode(Mapping())
        schema.add(SchemaNode(String(), name='login', validator=Length(min=2)))
        schema.add(SchemaNode(Boolean(), name='remember'))

        form = Form(
            DummyRequest(POST={'login': 'user1'}), schema=schema)
        self.assertFalse(form.validate())

        form = Form(
            DummyRequest(POST={'login': 'u', 'remember': False}),
            schema=schema)
        self.assertFalse(form.validate())

    # -------------------------------------------------------------------------
    def test_validate_ok(self):
        """[u:lib.form.Form.validate] OK"""
        from pyramid.testing import DummyRequest
        from ..lib.form import Form

        form = Form(DummyRequest(POST={'login': 'user1'}))
        form.validate()
        self.assertTrue(form.validate())
        self.assertNotIn('remember', form.values)
        self.assertIn('login', form.values)
        self.assertEqual(form.values['login'], 'user1')

        form = Form(
            DummyRequest(POST={'login': 'user1'}), schema=self.schema())
        self.assertTrue(form.validate())
        self.assertIn('remember', form.values)
        self.assertFalse(form.values['remember'])
        self.assertIn('login', form.values)
        self.assertEqual(form.values['login'], 'user1')

        class User(object):
            """Dummy user class."""
            # pylint: disable = too-few-public-methods
            login = None
            remember = False

        form = Form(
            DummyRequest(POST={'login': 'user1', 'remember': True}),
            schema=self.schema())
        user = User()
        self.assertTrue(form.validate(user))
        self.assertIn('remember', form.values)
        self.assertTrue(form.values['remember'])
        self.assertEqual(user.login, 'user1')
        self.assertTrue(user.remember)

    # -------------------------------------------------------------------------
    def test_has_error(self):
        """[u:lib.form.Form.has_error]"""
        from pyramid.testing import DummyRequest
        from ..lib.form import Form

        form = Form(
            DummyRequest(POST={'remember': True}), schema=self.schema())
        form.validate()
        self.assertTrue(form.has_error())
        self.assertTrue(form.has_error('login'))
        self.assertFalse(form.has_error('remember'))

    # -------------------------------------------------------------------------
    def test_set_error(self):
        """[u:lib.form.Form.set_error]"""
        from pyramid.testing import DummyRequest
        from ..lib.form import Form

        form = Form(DummyRequest())
        form.set_error('login', 'required')
        self.assertTrue(form.has_error('login'))

        form = Form(
            DummyRequest(POST={'remember': True}), schema=self.schema())
        form.validate()
        form.set_error('login', 'required login')
        self.assertTrue(form.has_error('login'))

    # -------------------------------------------------------------------------
    def test_error(self):
        """[u:lib.form.Form.error]"""
        from pyramid.testing import DummyRequest
        from ..lib.form import Form

        form = Form(
            DummyRequest(POST={'remember': True}), schema=self.schema())
        form.validate()
        self.assertEqual(form.error('login'), 'Required')
        self.assertFalse(form.error('remember'))

    # -------------------------------------------------------------------------
    def test_static(self):
        """[u:lib.form.Form.static]"""
        from pyramid.testing import DummyRequest
        from ..lib.form import Form

        form = Form(
            DummyRequest(POST={'login': 'user2'}),
            schema=self.schema(), defaults={'login': 'user1'})
        form.static('login')
        input_text = str(form.text('login', 'user1'))
        self.assertIn('user1', input_text)
        self.assertNotIn('user2', input_text)

    # -------------------------------------------------------------------------
    def test_forget(self):
        """[u:lib.form.Form.forget]"""
        from pyramid.testing import DummyRequest
        from ..lib.form import Form

        form = Form(
            DummyRequest(POST={'login': 'user1'}), schema=self.schema())
        form.forget('l')
        input_text = str(form.text('login'))
        self.assertNotIn('user1', input_text)

    # -------------------------------------------------------------------------
    def test_make_safe_id(self):
        """[u:lib.form.Form.make_safe_id]"""
        from ..lib.form import Form

        self.assertEqual(Form.make_safe_id('~Publidoc'), 'publidoc')
        self.assertEqual(Form.make_safe_id('#12'), '12')

    # -------------------------------------------------------------------------
    def test_begin(self):
        """[u:lib.form.Form.begin]"""
        from pyramid.csrf import new_csrf_token
        from pyramid.testing import DummyRequest
        from ..lib.form import Form

        form = Form(DummyRequest())
        html = str(form.begin(url='/login', multipart=True))
        self.assertTrue(html.startswith(
            '<form action="/login" enctype="multipart/form-data"'
            ' method="post">'))

        request = DummyRequest()
        token = new_csrf_token(request)
        form = Form(request)
        html = str(form.begin(url='/login'))
        self.assertEqual(
            html,
            '<form action="/login" method="post"><div class="cioHidden">'
            '<input id="csrf_token" name="csrf_token" type="hidden" '
            'value="{0}" /></div>'.format(token))

    # -------------------------------------------------------------------------
    def test_end(self):
        """[u:lib.form.Form.end]"""
        from pyramid.testing import DummyRequest
        from ..lib.form import Form

        html = str(Form(DummyRequest()).end())
        self.assertEqual(html, '</form>')

    # -------------------------------------------------------------------------
    def test_submit(self):
        """[u:lib.form.Form.submit]"""
        from pyramid.testing import DummyRequest
        from ..lib.form import Form

        html = Form(DummyRequest()).submit('foo', 'Foo')
        self.assertIn('id="foo"', html)
        self.assertIn('value="Foo"', html)
        self.assertIn('class="cioButton"', html)

        html = Form(DummyRequest()).submit('foo', class_='mainButton')
        self.assertIn('class="mainButton"', html)

    # -------------------------------------------------------------------------
    def test_submit_image(self):
        """[u:lib.form.Form.submit_image]"""
        from pyramid.testing import DummyRequest
        from ..lib.form import Form

        html = Form(DummyRequest()).submit_image('foo', '', 'Images/foo.png')
        self.assertIn('type="image"', html)
        self.assertIn('name="foo"', html)
        self.assertIn('title="foo"', html)

        html = Form(DummyRequest()).submit_image(
            'foo', 'My "Foo"', 'Images/foo.png')
        self.assertIn('title="My &#34;Foo&#34;', html)

        html = Form(DummyRequest()).submit_image(
            'foo', 'éÔ', 'Images/foo.png')
        self.assertIn('title="éÔ"', html)

    # -------------------------------------------------------------------------
    def test_submit_cancel(self):
        """[u:lib.form.Form.submit_cancel]"""
        from pyramid.testing import DummyRequest
        from ..lib.form import Form

        html = Form(DummyRequest()).submit_cancel(
            'Cancel', 'Images/cancel.png')
        self.assertIn('type="image"', html)
        self.assertIn('name="ccl!"', html)
        self.assertIn('title="Cancel"', html)

    # -------------------------------------------------------------------------
    def test_button(self):
        """[u:lib.form.Form.button]"""
        from pyramid.testing import DummyRequest
        from ..lib.form import Form

        html = Form(DummyRequest()).button(
            'http://localhost/foo', 'My Foo')
        self.assertIn('href="http://localhost/foo"', html)
        self.assertIn('class="cioButton"', html)

    # -------------------------------------------------------------------------
    def test_default_button(self):
        """[u:lib.form.Form.default_button]"""
        from pyramid.testing import DummyRequest
        from ..lib.form import Form

        html = Form(DummyRequest()).default_button('filter')
        self.assertIn('name="filter"', html)

    # -------------------------------------------------------------------------
    def test_grid_item(self):
        """[u:lib.form.Form.grid_item]"""
        from webhelpers2.html import literal
        from pyramid.testing import DummyRequest
        from ..lib.form import Form

        html = Form(DummyRequest()).grid_item(
            'My label', literal('<input type="test" id="foo">'),
            required=True, hint='éÔ', error='ôÉ')
        self.assertIn('<em> éÔ</em>', html)
        self.assertIn('<strong> ôÉ</strong>', html)
        self.assertIn('<span>*</span>', html)

    # -------------------------------------------------------------------------
    def test_hidden(self):
        """[u:lib.form.Form.hidden]"""
        from pyramid.testing import DummyRequest
        from ..lib.form import Form

        html = Form(DummyRequest()).hidden('foo', 'éÔ')
        self.assertIn('type="hidden"', html)
        self.assertIn('id="foo"', html)
        self.assertIn('value="éÔ"', html)

    # -------------------------------------------------------------------------
    def test_text(self):
        """[u:lib.form.Form.text]"""
        from pyramid.testing import DummyRequest
        from ..lib.form import Form

        html = Form(DummyRequest()).text('foo', 'éÔ')
        self.assertIn('type="text"', html)
        self.assertIn('id="foo"', html)
        self.assertIn('value="éÔ"', html)

    # -------------------------------------------------------------------------
    def test_password(self):
        """[u:lib.form.Form.password]"""
        from pyramid.testing import DummyRequest
        from ..lib.form import Form

        html = Form(DummyRequest()).password('password', 'seekrit')
        self.assertIn('type="password"', html)
        self.assertIn('id="password"', html)
        self.assertIn('value="seekrit"', html)

    # -------------------------------------------------------------------------
    def test_checkbox(self):
        """[u:lib.form.Form.checkbox]"""
        from pyramid.testing import DummyRequest
        from ..lib.form import Form

        html = Form(DummyRequest()).checkbox('foo', 'éÔ')
        self.assertIn('type="checkbox"', html)
        self.assertIn('value="éÔ"', html)
        self.assertNotIn('checked="checked"', html)

        html = Form(DummyRequest()).checkbox('foo', checked=True)
        self.assertIn('checked="checked"', html)

        html = Form(DummyRequest()).checkbox('foo', autosubmit=True)
        self.assertIn('class="cioAutoSubmit"', html)

        html = Form(DummyRequest(POST={'foo': '1'})).checkbox('foo')
        self.assertIn('checked="checked"', html)

    # -------------------------------------------------------------------------
    def test_custom_checkbox(self):
        """[u:lib.form.Form.custom_checkbox]"""
        from pyramid.testing import DummyRequest
        from ..lib.form import Form

        html = Form(DummyRequest()).custom_checkbox('foo')
        self.assertIn('cioCustomCheckbox', html)
        self.assertIn('label', html)
        self.assertIn('type="checkbox"', html)
        self.assertIn('value="1"', html)
        self.assertNotIn('checked="checked"', html)

        html = Form(DummyRequest()).custom_checkbox('foo', checked=True)
        self.assertIn('checked="checked"', html)

        html = Form(DummyRequest()).custom_checkbox('foo', autosubmit=True)
        self.assertIn('class="cioCustomCheckbox cioAutoSubmit"', html)

        html = Form(DummyRequest(POST={'foo': '1'})).custom_checkbox('foo')
        self.assertIn('checked="checked"', html)

    # -------------------------------------------------------------------------
    def test_radio(self):
        """[u:lib.form.Form.radio]"""
        from pyramid.testing import DummyRequest
        from ..lib.form import Form

        html = Form(DummyRequest()).radio('foo', 'éÔ')
        self.assertIn('type="radio"', html)
        self.assertIn('value="éÔ"', html)
        self.assertNotIn('checked="checked"', html)

        html = Form(DummyRequest()).radio('foo', 'bar', checked=True)
        self.assertIn('checked="checked"', html)

        html = Form(DummyRequest()).radio('foo', 'bar', autosubmit=True)
        self.assertIn('class="cioAutoSubmit"', html)

        html = Form(DummyRequest(POST={'foo': 'éÔ'})).radio('foo', 'éÔ')
        self.assertIn('checked="checked"', html)

    # -------------------------------------------------------------------------
    def test_select(self):
        """[u:lib.form.Form.select]"""
        from pyramid.testing import DummyRequest
        from ..lib.form import Form

        html = Form(DummyRequest()).select('foo', 'b', None)
        self.assertEqual(html, '')

        html = Form(DummyRequest()).select(
            'foo', 'b', ('', 'a', 'b', 'é'), autosubmit=True)
        self.assertIn('<select', html)
        self.assertIn('id="foo"', html)
        self.assertIn('class="cioAutoSubmit"', html)
        self.assertIn('<option></option>', html)
        self.assertIn('<option>a</option>', html)
        self.assertIn('<option selected="selected">b</option>', html)
        self.assertIn('<option>é</option>', html)

        html = Form(DummyRequest()).select('foo', None, (1, 2))
        self.assertIn('<option>1</option>', html)
        self.assertIn('<option>2</option>', html)

        html = Form(DummyRequest()).select(
            'foo', 'b', (('', ''), ('a', 'A'), ('b', 'B'), ('é', 'É')))
        self.assertIn('<option value="a">A</option>', html)
        self.assertIn('<option selected="selected" value="b">B</option>', html)
        self.assertIn('<option value="é">É</option>', html)

        html = Form(DummyRequest()).select(
            'foo', None, ((1, 'One'), (2, 'Two')))
        self.assertIn('<option value="1">One</option>', html)

    # -------------------------------------------------------------------------
    def test_upload(self):
        """[u:lib.form.Form.upload]"""
        from pyramid.testing import DummyRequest
        from ..lib.form import Form

        html = Form(DummyRequest()).upload('upload_file')
        self.assertIn('type="file"', html)
        self.assertIn('id="upload_file"', html)

    # -------------------------------------------------------------------------
    def test_textarea(self):
        """[u:lib.form.Form.textarea]"""
        from pyramid.testing import DummyRequest
        from ..lib.form import Form

        html = Form(DummyRequest()).textarea('foo')
        self.assertIn('<textarea', html)
        self.assertIn('id="foo"', html)

    # -------------------------------------------------------------------------
    def test_grid_text(self):
        """[u:lib.form.Form.grid_text]"""
        from pyramid.testing import DummyRequest
        from ..lib.form import Form

        html = Form(DummyRequest()).grid_text('foo', 'éÔ', clear=True)
        self.assertIn('<div class="cioFormItem">', html)
        self.assertIn('<label for="foo">', html)
        self.assertIn('type="text"', html)
        self.assertIn('<strong>éÔ</strong>', html)
        self.assertIn('<div class="cioClear">', html)

    # -------------------------------------------------------------------------
    def test_grid_password(self):
        """[u:lib.form.Form.grid_password]"""
        from pyramid.testing import DummyRequest
        from ..lib.form import Form

        html = Form(DummyRequest()).grid_password('password', 'Password')
        self.assertIn('<div class="cioFormItem">', html)
        self.assertIn('<label for="password">', html)
        self.assertIn('type="password"', html)
        self.assertIn('<strong>Password</strong>', html)
        self.assertNotIn('<div class="cioClear">', html)

    # -------------------------------------------------------------------------
    def test_grid_checkbox(self):
        """[u:lib.form.Form.grid_checkbox]"""
        from pyramid.testing import DummyRequest
        from ..lib.form import Form

        html = Form(DummyRequest()).grid_checkbox('foo', 'éÔ')
        self.assertIn('<div class="cioFormItem">', html)
        self.assertIn('<label for="foo">', html)
        self.assertIn('type="checkbox"', html)
        self.assertIn('<label for="foo">', html)

    # -------------------------------------------------------------------------
    def test_grid_custom_checkbox(self):
        """[u:lib.form.Form.grid_checkbox]"""
        from pyramid.testing import DummyRequest
        from ..lib.form import Form

        html = Form(DummyRequest()).grid_custom_checkbox('foo', 'éÔ')
        self.assertIn('<div class="cioFormItem">', html)
        self.assertIn('cioCustomCheckbox', html)

    # -------------------------------------------------------------------------
    def test_grid_select(self):
        """[u:lib.form.Form.grid_select]"""
        from pyramid.testing import DummyRequest
        from ..lib.form import Form

        html = Form(DummyRequest()).grid_select('foo', 'b', None)
        self.assertEqual(html, '')

        html = Form(DummyRequest()).grid_select(
            'foo', 'éÔ', ('', 'a', 'b', 'é'), autosubmit=True, required=True,
            hint='Ôé', clear=True)
        self.assertIn('<div class="cioFormItem">', html)
        self.assertIn('<label for="foo">', html)
        self.assertIn('<select', html)
        self.assertIn('<span>*</span>', html)
        self.assertIn('<div class="cioClear">', html)

    # -------------------------------------------------------------------------
    def test_grid_upload(self):
        """[u:lib.form.Form.grid_upload]"""
        from pyramid.testing import DummyRequest
        from ..lib.form import Form

        html = Form(DummyRequest()).grid_upload('foo', 'éÔ')
        self.assertIn('<div class="cioFormItem">', html)
        self.assertIn('<label for="foo">', html)
        self.assertIn('type="file"', html)

    # -------------------------------------------------------------------------
    def test_grid_textarea(self):
        """[u:lib.form.Form.grid_texarea]"""
        from pyramid.testing import DummyRequest
        from ..lib.form import Form

        html = Form(DummyRequest()).grid_textarea('foo', 'éÔ')
        self.assertIn('<div class="cioFormItem">', html)
        self.assertIn('<label for="foo">', html)
        self.assertIn('<textarea', html)
        self.assertIn('id="foo"', html)

    # -------------------------------------------------------------------------
    def test_values(self):
        """[u:lib.form.Form.values]"""
        from pyramid.testing import DummyRequest
        from ..lib.form import Form

        request = DummyRequest()
        request.POST['foo'] = 'éÔ'
        html = Form(request).text('foo')
        self.assertIn('value="éÔ"', html)

        form = Form(request)
        form.values['bar'] = 'ôÉ'
        html = form.text('bar')
        self.assertIn('value="ôÉ"', html)



# pylint: disable = import-outside-toplevel
"""Tests of ``modules.cioskeleton`` classes and functions."""

from unittest import TestCase

from ....tests import ModuleTestCase


# =============================================================================
class UCioSkeletonIncludeme(ModuleTestCase):
    """Unit test class for :func:`modules.cioskeleton.includeme`."""

    # -------------------------------------------------------------------------
    def test_configurator(self):
        """[u:modules.cioskeleton.includeme] with a configurator"""
        from ....tests import TEST_INI
        from .. import includeme, ModuleCioSkeleton

        self.configurator.get_settings()['__file__'] = TEST_INI
        self.configurator.registry['themes'] = {
            '': {'path': None, 'name': {}, 'layout': 'layout.pt'}}

        includeme(self.configurator)
        self.assertTrue(self.configurator.registry['modules'])
        self.assertIn(
            'chrysalio.modules.cioskeleton',
            self.configurator.registry['modules'])
        self.assertIsInstance(
            self.configurator.registry['modules'][
                'chrysalio.modules.cioskeleton'],
            ModuleCioSkeleton)

    # -------------------------------------------------------------------------
    def test_dictionary(self):
        """[u:modules.cioskeleton.includeme] with a dictionary"""
        from collections import OrderedDict
        from ....tests import TEST_INI
        from ....security import PRINCIPALS
        from ....scripts import ScriptRegistry
        from .. import includeme, ModuleCioSkeleton

        registry = ScriptRegistry({'__file__': TEST_INI})
        registry['modules'] = OrderedDict()
        registry['principals'] = list(PRINCIPALS)

        includeme(registry)
        self.assertEqual(len(registry['modules']), 1)
        self.assertIn('chrysalio.modules.cioskeleton', registry['modules'])
        self.assertIsInstance(
            registry['modules']['chrysalio.modules.cioskeleton'],
            ModuleCioSkeleton)


# =============================================================================
class UModuleCioSkeleton(TestCase):
    """Unit test class for :class:`modules.cioskeleton.ModuleCioSkeleton`."""

    # -------------------------------------------------------------------------
    def test_activate(self):
        """[u:modules.cioskeleton.ModuleCioSkeleton.activate]"""
        from ....tests import TEST_INI
        from ....security import PRINCIPALS
        from ....menu import MENU_HOME, MENU_ADMIN
        from ....modes import MODE_HOME
        from ..security import PRINCIPALS_CIOSKELETON
        from ..menu import MENU_CIOSKELETON
        from ..modes import MODE_CIOSKELETON
        from .. import ModuleCioSkeleton

        registry = {
            'principals': list(PRINCIPALS), 'menu': [MENU_HOME, MENU_ADMIN]}
        ModuleCioSkeleton(TEST_INI).activate(registry, None)
        self.assertIn(PRINCIPALS_CIOSKELETON[1], registry['principals'])
        self.assertIn(MENU_CIOSKELETON, registry['menu'])

        registry['modes'] = [MODE_HOME]
        ModuleCioSkeleton(TEST_INI).activate(registry, None)
        self.assertIn(MODE_CIOSKELETON, registry['modes'])

    # -------------------------------------------------------------------------
    def test_deactivate(self):
        """[u:modules.cioskeleton.ModuleCioSkeleton.deactivate]"""
        from ....tests import TEST_INI
        from ....security import PRINCIPALS
        from ....modes import MODE_HOME
        from ....menu import MENU_HOME, MENU_ADMIN
        from ..security import PRINCIPALS_CIOSKELETON
        from ..modes import MODE_CIOSKELETON
        from ..menu import MENU_CIOSKELETON
        from .. import ModuleCioSkeleton

        registry = {
            'principals': list(PRINCIPALS), 'menu': [MENU_HOME, MENU_ADMIN]}
        registry['menu'].insert(-1, MENU_CIOSKELETON)
        ModuleCioSkeleton(TEST_INI).deactivate(registry, None)
        self.assertNotIn(PRINCIPALS_CIOSKELETON[1], registry['principals'])
        self.assertNotIn(MENU_CIOSKELETON, registry['menu'])

        registry['modes'] = [MODE_HOME, MODE_CIOSKELETON]
        registry['principals'].append(PRINCIPALS_CIOSKELETON[1])
        ModuleCioSkeleton(TEST_INI).deactivate(registry, None)
        self.assertNotIn(PRINCIPALS_CIOSKELETON[1], registry['principals'])
        self.assertNotIn(MODE_CIOSKELETON, registry['modes'])



"""Collection of functionalities with an `includeme()` function to be used by
the Pyramid configurator or in the INI file."""

from sys import exit as sys_exit

from transaction import manager

from ..lib.i18n import _, translate
from ..lib.config import settings_get_list
from ..modules import Module
from ..models import get_tm_dbsession


# =============================================================================
def load_includes(configurator, includes=None):
    """Load modules with an `includeme()` function and fill
    ``configurator.registry['modules']`` ordered dictionary.

    :type  configurator: pyramid.config.Configurator
    :param configurator:
        Object used to do configuration declaration within the application.
    :param list includes: (optional)
        List of hard coding `includes`.
    """
    # Load includes and modules
    settings = configurator.get_settings()
    if includes is None:
        includes = []
    includes += [
        k for k in settings_get_list(settings, 'chrysalio.includes') if k]
    for include in includes:
        try:
            configurator.include(include)
        except ImportError as error:
            sys_exit(translate(
                _('*** Unable to load module "${m}" [${e}]', {
                    'm': include, 'e': error})))

    # Check modules
    if configurator.registry.get('modules'):
        # Check conflicts
        modules = configurator.registry['modules']
        implementations, error = Module.check_conflicts(includes, modules)
        if error is not None:
            sys_exit(translate(error))

        # Check dependencies
        for module_id in modules:
            error = modules[module_id].check_dependencies(implementations)
            if error is not None:
                sys_exit(translate(error))

        # Check activations
        activation = True
        while activation:
            activation = False
            for module_id in modules:
                activation |= modules[module_id].check_activations(
                    configurator.registry['modules_off'])

        # Activate modules
        with manager:
            dbsession = get_tm_dbsession(
                configurator.registry['dbsession_factory'], manager)
            for module_id in modules:
                if module_id not in configurator.registry['modules_off']:
                    modules[module_id].activate(
                        configurator.registry, dbsession)



"""CioSkeleton route definitions."""


# =============================================================================
def includeme(configurator):
    """Function to include routes.

    :type  configurator: pyramid.config.Configurator
    :param configurator:
        Object used to do configuration declaration within the application.
    """
    configurator.add_route('bone_index', '/bone/index')



"""Subscriber definition."""

from os.path import splitext

from pyramid.events import BeforeRender
from pyramid.threadlocal import get_current_request
from pyramid.i18n import TranslationString

from .lib.utils import tostr, camel_case
from .lib.breadcrumbs import Breadcrumbs
from .lib.panel import Panel
from .includes.themes import theme_template, theme_static_prefix
from .includes.themes import theme_has_static


# =============================================================================
def includeme(configurator):
    """Function to include subscribers.

    :type  configurator: pyramid.config.Configurator
    :param configurator:
        Object used to do configuration declaration within the application.
    """
    configurator.add_subscriber(before_render, BeforeRender)
    configurator.add_request_method(Breadcrumbs, 'breadcrumbs', reify=True)


# =============================================================================
def before_render(event):
    """A subscriber for :class:`pyramid.events.BeforeRender` events.

    :type  event: pyramid.events.BeforeRender
    :param event:
        Current event.
    """
    request = event.get('request') or get_current_request()
    domain = event.get('renderer_name', '')
    domain = domain.partition(':')[0] if ':' in domain else 'chrysalio'

    def translate(text, mapping=None, domain=domain):
        """Translation from a string."""
        return request.localizer.translate(TranslationString(
            text, mapping=mapping, domain=domain))

    event['_'] = translate
    event['route'] = lambda name, *elts, **kwargs: \
        tostr(request.route_path(name, *elts, **kwargs))
    event['title'] = request.registry['settings']['title']
    event['page_id'] = camel_case(
        splitext(event.get('renderer_name', '').split('/')[-1:][0])[0])
    event['theme'] = theme_static_prefix(request)
    event['theme_has'] = theme_has_static
    event['panel_css'] = Panel.open_panel_css(request)
    event['panel_js'] = Panel.open_panel_js(request)
    event['layout'] = theme_template(request, 'layout')



# -*- coding: utf-8 -*-
# pylint: disable = import-outside-toplevel
"""Tests of ``models.dbgroup`` classes."""

from json import dumps, loads
from collections import namedtuple

from lxml import etree
from colander import SchemaNode

from pyramid.i18n import TranslationString

from ..models.dbgroup import DBGroup
from . import DBUnitTestCase


# =============================================================================
class UModelsDBGroupDBGroup(DBUnitTestCase):
    """Unit test class for :class:`models.dbgroup.DBGroup`."""

    # -------------------------------------------------------------------------
    def test_xml2db(self):
        """[u:.models.dbgroup.DBGroup.xml2db]"""
        from ..models.dbuser import DBUser
        from ..models.dbgroup import DBGroupProfile

        dbsession = self.request.dbsession
        self.add_user({
            'login': 'member1', 'first_name': 'John', 'last_name': 'DEUF',
            'password': 'member1pwd', 'email': 'member1@chrysal.io'})
        dbsession.add(DBGroup(
            group_id='managers', i18n_label=dumps({'en': 'Managers'})))

        # error_if_exists
        group_elt = etree.XML(
            '<group id="managers">'
            '  <label xml:lang="en">Managers</label>'
            '</group>')
        error = DBGroup.xml2db(dbsession, group_elt)
        self.assertIsNotNone(error)
        error = DBGroup.xml2db(dbsession, group_elt, error_if_exists=False)
        self.assertIsNone(error)

        # Without content
        group_elt = etree.XML(
            '<group id="Foo">'
            '</group>')
        error = DBGroup.xml2db(dbsession, group_elt)
        self.assertIsNotNone(error)

        # A valid group
        group_elt = etree.XML(
            '<group id="editors">'
            '  <label xml:lang="en">Editors</label>'
            '  <description xml:lang="en">Éditeurs</description>'
            '  <attachments key="Editors">'
            '    <picture>editors.png</picture>'
            '  </attachments>'
            '  <users>'
            '    <user>member1</user>'
            '  </users>'
            '  <profiles>'
            '    <profile>profile_editor</profile>'
            '    <profile>user_editor</profile>'
            '  </profiles>'
            '</group>')
        user_id = dbsession.query(DBUser.user_id).filter_by(
            login='member1').first()
        error = DBGroup.xml2db(dbsession, group_elt, kwargs={
            'users': {'member1': user_id[0]},
            'profiles': ('profile_editor', 'user_editor')})
        self.assertIsNone(error)
        dbgroup = dbsession.query(DBGroup).filter_by(
            group_id='editors').first()
        self.assertIsInstance(dbgroup, DBGroup)
        self.assertNotIsInstance(dbgroup.i18n_label, dict)
        self.assertIsInstance(dbgroup.i18n_description, dict)
        self.assertIn('en', dbgroup.i18n_description)
        self.assertEqual(len(dbgroup.users), 1)
        self.assertEqual(len(dbsession.query(DBGroupProfile).all()), 2)

    # -------------------------------------------------------------------------
    def test_record_from_xml(self):
        """[u:models.dbgroup.DBGroup.record_from_xml]"""
        group_elt = etree.XML(
            '<group id="editors">'
            '  <label xml:lang="en">Editors</label>'
            '  <attachments key="Editors">'
            '    <picture>editors.png</picture>'
            '  </attachments>'
            '  <users>'
            '    <user>member1</user>'
            '  </users>'
            '  <profiles>'
            '    <profile>profile_editor</profile>'
            '    <profile>group_editor</profile>'
            '  </profiles>'
            '</group>')
        record = DBGroup.record_from_xml('editors', group_elt)
        self.assertEqual(record['group_id'], 'editors')
        self.assertIn('i18n_label', record)
        labels = loads(record['i18n_label'])
        self.assertIn('en', labels)
        self.assertEqual(labels['en'], 'Editors')

    # -------------------------------------------------------------------------
    def test_record_format(self):
        """[u:models.dbgroup.DBGroup.record_format]"""
        record = {'group_id': None}
        error = DBGroup.record_format(record)
        self.assertIsInstance(error, TranslationString)
        self.assertIn('without ID', error)

        record = {'group_id': 'team1'}
        error = DBGroup.record_format(record)
        self.assertIsInstance(error, TranslationString)
        self.assertIn('without label', error)

        record = {
            'group_id': 'team1', 'label_en': 'Team 1',
            'label_fr': 'Équipe 1', 'description_en': 'Team of users n° 1.'}
        error = DBGroup.record_format(record)
        self.assertIsNone(error)
        self.assertIn('i18n_label', record)
        labels = loads(record['i18n_label'])
        self.assertEqual(len(labels), 2)
        self.assertIn('fr', labels)
        self.assertEqual(labels['fr'], 'Équipe 1')
        self.assertIn('i18n_description', record)
        self.assertEqual(len(record['i18n_description']), 1)

        record = {
            'group_id': 'team1', 'label_en': 'Team 1', 'label_fr': 'Équipe 1'}
        error = DBGroup.record_format(record)
        self.assertIsNone(error)
        self.assertIn('i18n_description', record)

    # -------------------------------------------------------------------------
    def test_db2xml(self):
        """[u:models.dbgroup.DBGroup.db2xml]"""
        from ..models.dbprofile import DBProfile
        from ..models.dbuser import DBUser
        from ..models.dbgroup import DBGroupUser

        dbsession = self.request.dbsession
        self.add_user({
            'login': 'member1', 'first_name': 'John', 'last_name': 'DEUF',
            'password': 'member1pwd', 'email': 'member1@chrysal.io'})
        user_id = dbsession.query(DBUser.user_id).filter_by(
            login='member1').first()[0]
        dbgroup = DBGroup(
            group_id='managers',
            i18n_label=dumps({'en': 'Managers'}),
            i18n_description={'en': 'Permission to view, edit and create.'},
            attachments_key='Managers', picture='managers.png')
        dbgroup.users.append(DBGroupUser(user_id=user_id))
        dbgroup.profiles.append(DBProfile(profile_id='user_manager'))

        group_elt = dbgroup.db2xml(dbsession)
        self.assertEqual(group_elt.get('id'), 'managers')
        self.assertEqual(group_elt.findtext('label'), 'Managers')
        self.assertEqual(
            group_elt.findtext('description').strip(),
            'Permission to view, edit and create.')
        self.assertIsNotNone(group_elt.find('users'))
        self.assertEqual(group_elt.findtext('users/user'), 'member1')
        self.assertIsNotNone(group_elt.find('profiles'))
        self.assertEqual(
            group_elt.findtext('profiles/profile'), 'user_manager')

    # -------------------------------------------------------------------------
    def test_tab4view(self):
        """[u:models.dbgroup.DBGroup.tab4view]"""
        from ..lib.form import Form
        from ..lib.paging import Paging
        from ..lib.filter import Filter
        from ..models.dbuser import DBUser
        from ..models.dbprofile import DBProfile

        dbsession = self.request.dbsession
        form = Form(self.request)
        paging_id = 'group_users'
        user_filter = Filter(
            self.request, paging_id, (('login', 'Login', False, None),))
        user_paging = Paging(self.request, paging_id, [
            DBUser(user_id=1, login='user1', last_name='Mr TEST')])
        dbgroup = DBGroup(
            group_id='viewers',
            i18n_label=dumps({'en': 'Observers', 'fr': 'Observateurs'}),
            i18n_description={'en': 'These users do not touch.'},
            attachments_key='Viewers', picture='viewers.png')
        dbsession.add(dbgroup)
        self.configurator.add_route('home', '/')
        self.configurator.add_route(
            'profile_view', '/profile/view/{profile_id}')
        self.configurator.add_route('user_view', '/user/view/{user_id}')
        self.request.matched_route = namedtuple('Route', 'name')(name='home')

        # Information
        html = dbgroup.tab4view(
            self.request, 0, form, user_filter, user_paging)
        self.assertIn('viewers', html)
        self.assertIn('These users do not touch.', html)

        # Users
        html = dbgroup.tab4view(
            self.request, 1, form, user_filter, user_paging)
        self.assertIn('Mr TEST', html)

        # Profiles
        html = dbgroup.tab4view(
            self.request, 2, form, user_filter, user_paging)
        self.assertIn('No attributed profile.', html)
        dbgroup.profiles.append(DBProfile(
            profile_id='user_viewer', i18n_label=dumps({'en': 'User Viewer'})))
        html = dbgroup.tab4view(
            self.request, 2, form, user_filter, user_paging)
        self.assertIn('User Viewer', html)

        # Other
        self.assertEqual(
            dbgroup.tab4view(self.request, 3, form, user_filter, user_paging),
            '')

    # -------------------------------------------------------------------------
    def test_settings_schema(self):
        """[u:models.dbgroup.DBGroup.settings_schema]"""
        from ..models.dbprofile import DBProfile
        from ..models.dbgroup import DBGroupUser

        self.request.registry.settings['languages'] = 'en, fr'
        self.add_user({
            'login': 'member1', 'first_name': 'John', 'last_name': 'DEUF',
            'password': 'member1pwd', 'email': 'member1@chrysal.io'})

        # Creation
        schema, defaults = DBGroup.settings_schema(
            self.request, {}, ('profile_manager', 'user_manager'))
        self.assertIsInstance(schema, SchemaNode)
        serialized = schema.serialize()
        self.assertIn('group_id', serialized)
        self.assertIn('label_en', serialized)
        self.assertIn('label_fr', serialized)
        self.assertIn('description_en', serialized)
        self.assertIn('description_fr', serialized)
        self.assertIn('pfl:user_manager', serialized)
        self.assertIsInstance(defaults, dict)
        self.assertFalse(defaults)

        # Update
        dbgroup = DBGroup(
            group_id='managers',
            i18n_label=dumps({'en': 'Managers', 'fr': 'Gestionnaires'}),
            i18n_description={'en': 'These users manage.'},
            attachments_key='Managers', picture='managers.png')
        dbgroup.users.append(DBGroupUser(user_id='member1'))
        dbgroup.profiles.append(DBProfile(profile_id='user_manager'))
        schema, defaults = DBGroup.settings_schema(
            self.request, {}, ('profile_manager', 'user_manager'), dbgroup)
        self.assertNotIn('group_id', schema.serialize())
        self.assertIn('label_en', defaults)
        self.assertIn('description_en', defaults)
        self.assertIn('pfl:user_manager', defaults)

    # -------------------------------------------------------------------------
    def test_tab4edit(self):
        """[u:models.dbgroup.DBGroup.tab4edit]"""
        from ..lib.form import Form
        from ..lib.paging import Paging
        from ..lib.filter import Filter
        from ..models.dbuser import DBUser

        paging_id = 'group_users'
        self.request.registry.settings['languages'] = 'en, fr'
        user_filter = Filter(
            self.request, paging_id, (('login', 'Login', False, None),))
        user_paging = Paging(self.request, paging_id, [
            DBUser(user_id=1, login='test1', last_name='Mr TEST')])
        form = Form(self.request)
        dbgroup = DBGroup(
            group_id='managers',
            i18n_label=dumps({'en': 'Managers', 'fr': 'Gestionnaires'}),
            i18n_description={'en': 'These users manage.'},
            attachments_key='Managers', picture='managers.png')
        profiles = {
            'user_creator': ('User Creator', 'Authorized to create users.'),
            'user_editor': ('User Editor', None)}
        self.configurator.add_route('home', '/')
        self.request.matched_route = namedtuple('Route', 'name')(name='home')

        # Information, creation
        html = DBGroup.tab4edit(
            self.request, 0, form, user_filter, user_paging, {})
        self.assertIn('Identifier:', html)
        self.assertIn('group_id', html)
        self.assertIn('label_en', html)
        self.assertIn('label_fr', html)
        self.assertIn('description_en', html)
        self.assertIn('description_fr', html)

        # Information, update
        html = DBGroup.tab4edit(
            self.request, 0, form, user_filter, user_paging, profiles, dbgroup)
        self.assertNotIn('group_id', html)

        # Users
        html = DBGroup.tab4edit(
            self.request, 1, form, user_filter, user_paging, {}, dbgroup)
        self.assertIn('test1', html)

        # Profiles
        html = DBGroup.tab4edit(
            self.request, 2, form, user_filter, user_paging, {})
        self.assertEqual(html, 'No available profile.')
        html = DBGroup.tab4edit(
            self.request, 2, form, user_filter, user_paging, profiles, dbgroup)
        self.assertIn('User Creator', html)
        self.request.has_permission = lambda x: False
        html = DBGroup.tab4edit(
            self.request, 2, form, user_filter, user_paging, profiles, dbgroup)
        self.assertEqual(html, 'You do not have the rigths to edit profiles.')

        # Other
        html = DBGroup.tab4edit(
            self.request, 3, form, user_filter, user_paging, {})
        self.assertEqual(html, '')


# =============================================================================
class IModelsDBGroup(DBUnitTestCase):
    """Integration test class for ``models.dbgroup``."""

    # -------------------------------------------------------------------------
    def test_delete_cascade_user(self):
        """[i:models.dbgroup] delete, cascade user"""
        from ..models.dbuser import DBUser
        from ..models.dbgroup import DBGroupUser

        dbsession = self.request.dbsession
        self.add_user({
            'login': 'member2', 'first_name': 'Cécile',
            'last_name': 'OURKESSA', 'password': 'member2pwd',
            'email': 'member2@chrysal.io'})
        user_id = dbsession.query(DBUser.user_id).filter_by(
            login='member2').first()[0]
        group_id = 'managers'
        dbgroup = DBGroup(
            group_id=group_id, i18n_label=dumps({'en': 'Managers'}))
        dbgroup.users.append(
            DBGroupUser(user_id=user_id))
        dbsession.add(dbgroup)

        dbgroup = dbsession.query(DBGroup).filter_by(group_id=group_id).first()
        self.assertIsInstance(dbgroup, DBGroup)
        dbgroup_user = dbsession.query(DBGroupUser).filter_by(
            group_id=group_id).first()
        self.assertIsInstance(dbgroup_user, DBGroupUser)

        dbsession.delete(dbgroup)
        dbsession.flush()
        dbgroup = dbsession.query(DBGroup).filter_by(group_id=group_id).first()
        self.assertIsNone(dbgroup)
        dbgroup_user = dbsession.query(DBGroupUser).filter_by(
            group_id=group_id).first()
        self.assertIsNone(dbgroup_user)

    # -------------------------------------------------------------------------
    def test_delete_cascade_profile(self):
        """[i:models.dbgroup] delete, cascade group profile"""
        from ..models.dbprofile import DBProfile
        from ..models.dbgroup import DBGroupProfile

        dbsession = self.request.dbsession
        dbsession.add(DBProfile(
            profile_id='group_manager',
            i18n_label=dumps({'en': 'Group manager'})))
        group_id = 'editors'
        dbgroup = DBGroup(
            group_id=group_id, i18n_label=dumps({'en': 'Editors'}))
        dbsession.add(dbgroup)
        dbsession.flush()
        dbsession.add(DBGroupProfile(
            group_id=dbgroup.group_id, profile_id='group_manager'))

        dbgroup = dbsession.query(DBGroup).filter_by(group_id=group_id).first()
        self.assertIsInstance(dbgroup, DBGroup)
        self.assertIsNotNone(dbsession.query(DBGroupProfile).filter_by(
            group_id=group_id).first())

        dbgroup.profiles.remove(dbgroup.profiles[0])
        dbsession.flush()
        self.assertIsNone(dbsession.query(DBGroupProfile).filter_by(
            group_id=group_id).first())



# pylint: disable = import-outside-toplevel
"""Tests of ``views.panel`` functions."""

from collections import namedtuple
from unittest import TestCase

from pyramid.testing import DummyRequest, DummySession

from ..lib.panel import Panel


# =============================================================================
class DummyPanel(Panel):
    """A class to test panels."""

    uid = 'dummy'
    label = 'Side Panel'


# =============================================================================
class UViewsPanelPanelView(TestCase):
    """Unit test class for testing :func:`views.panel.panel_view`."""

    # -------------------------------------------------------------------------
    def test_it(self):
        """[u:views.panel.panel_view]"""
        from ..views.panel import panel_view

        request = DummyRequest()
        request.registry = {}
        request.session = DummySession({})
        panel = DummyPanel()

        # Without panels
        panel_view(request)
        self.assertNotIn('panels', request.registry)

        # With panels but an unknown panel to open
        request.registry = {'panels': {}}
        request.matchdict = {'panel_id': 'test'}
        panel_view(request)
        self.assertNotIn('test', request.registry['panels'])

        # With panels and a known panel to open
        request.matched_route = namedtuple('Route', 'name')(name='panel_open')
        request.registry['panels']['test'] = panel
        panel_view(request)
        self.assertFalse(panel.was_open(request))
        self.assertTrue(panel.is_open(request))

        # With panels and a known panel to close
        request.matched_route = namedtuple('Route', 'name')(name='panel_close')
        panel_view(request)
        self.assertFalse(panel.was_open(request))
        self.assertFalse(panel.is_open(request))



"""Function to import and export database from and into XML files."""

from ..relaxng import RELAXNG_CIOLDAP
from .dbldap import DBLdap


# =============================================================================
def xml2db(dbsession, root_elt, only=None, error_if_exists=True, modules=None):
    """Load an XML configuration file for an included module.

    :type  dbsession: sqlalchemy.orm.session.Session
    :param dbsession:
        SQLAlchemy session.
    :type  root_elt: lxml.etree.Element
    :param root_elt:
        XML element with the namespace of the module.
    :param str only: (optional)
        If not ``None``, only the items of type ``only`` are loaded.
    :param bool error_if_exists: (default=True)
        It returns an error if an item already exists.
    :type  modules: collections.OrderedDict
    :param modules: (optional)
        Dictionary of modules to use to complete the loading.
    :rtype: list
    :return:
        A list of error messages.
    """
    # pylint: disable = unused-argument
    if only is not None and only != 'ldap':
        return []

    ldap_elt = root_elt.find(
        '{{{0}}}ldap'.format(RELAXNG_CIOLDAP['namespace']))
    if ldap_elt is None:
        return []
    error = DBLdap.xml2db(dbsession, ldap_elt)
    return [error] if error is not None else []


# =============================================================================
def db2xml(dbsession, root_elt):
    """Return a list of XML elements.

    :type  dbsession: sqlalchemy.orm.session.Session
    :param dbsession:
        SQLAlchemy session.
    :type  root_elt: lxml.etree._Element
    :param root_elt:
        XML element with the namespace of the module.
    """
    dbldap = dbsession.query(DBLdap).first()
    if dbldap is not None:
        root_elt.append(dbldap.db2xml())



"""Security functionalities."""

from pyramid.authentication import AuthTktAuthenticationPolicy
from pyramid.authorization import ACLAuthorizationPolicy
from pyramid.security import Authenticated, Everyone, Allow, ALL_PERMISSIONS

from .lib.i18n import _
from .lib.log import log_info
from .models.dbuser import DBUser


PRINCIPALS = (
    ('mode', _('Mode management'), [
        ('admin', _('Access to administration mode'), ('mode-admin',)),
        ('job', _('Access to job mode'), ('mode-job',))
    ]),
    ('group', _('Group management'), (
        ('viewer', _('View all groups'), (
            'group-view',)),
        ('editor', _('Edit or view any group'), (
            'group-edit', 'group-view')),
        ('creator', _('Create a new one or edit or view any group'), (
            'group-create', 'group-edit', 'group-view'))
    )),
    ('user', _('User management'), (
        ('viewer', _('View all users'), (
            'user-view',)),
        ('editor', _('Edit or view any user'), (
            'user-edit', 'user-view')),
        ('creator', _('Create a new one or edit or view any user'), (
            'user-create', 'user-edit', 'user-view'))
    )),
    ('profile', _('Profile management'), (
        ('viewer', _('View any profile'), (
            'profile-view',)),
        ('editor', _('Edit or view any profile'), (
            'profile-edit', 'profile-view')),
        ('creator', _('Create a new one or edit or view any profile'), (
            'profile-create', 'profile-edit', 'profile-view'))
    )),
    ('settings', _('General settings management'), (
        ('viewer', _('View general settings'), (
            'settings-view',)),
        ('editor', _('Edit or view general settings'), (
            'settings-edit', 'settings-view'))
    )),
    ('backup', _('Backup management'), (
        ('creator', _('Backup and restore'), (
            'backup-create',)),
    ))
)


# =============================================================================
def includeme(configurator):
    """Function to include authentication mechanism.

    :type  configurator: pyramid.config.Configurator
    :param configurator:
        Object used to do configuration declaration within the application.
    """
    settings = configurator.get_settings()
    configurator.set_authorization_policy(ACLAuthorizationPolicy())
    configurator.set_authentication_policy(SessionAuthenticationPolicy(
        secret=settings.get('auth.secret', '-'), hashalg='sha512',
        cookie_name=settings.get('auth.cookie', 'CIO_AUTH'),
        secure=settings.get('auth.secure') == 'true',
        http_only=True, samesite=settings.get('auth.samesite', 'Strict')))
    configurator.set_root_factory(RootFactory)
    configurator.registry['principals'] = list(PRINCIPALS)


# =============================================================================
class SessionAuthenticationPolicy(AuthTktAuthenticationPolicy):
    """Authentication policy class based on session."""

    _DBUser = DBUser

    # -------------------------------------------------------------------------
    def authenticated_userid(self, request):
        """Return the authenticated :term:`userid` or ``None`` if
        no authenticated user ID can be found.

        :type  request: pyramid.request.Request
        :param request:
            Current request.
        :rtype: :class:`str` or ``None``
        """
        user = self._user(request)
        if user is not None:
            return user['user_id']
        return None

    # -------------------------------------------------------------------------
    def effective_principals(self, request):
        """Return a list of principals including, at least,
        :class:`pyramid.security.Everyone`.

        :type  request: pyramid.request.Request
        :param request:
            Current request.
        :return: (list)
        """
        principals = [Everyone]
        user = self._user(request)
        if user is not None:
            principals.append(Authenticated)
            principals += user['principals']
        return principals

    # -------------------------------------------------------------------------
    @classmethod
    def _user(cls, request):
        """Get user from session.

        :type  request: pyramid.request.Request
        :param request:
            Current request.
        :rtype: dict
        """
        # Already authenticated
        login = request.unauthenticated_userid
        if login and 'user' in request.session and \
           request.session['user']['login'] == login:
            return request.session['user']

        # Auto login
        dbuser = cls._DBUser.get(request, login)[0]
        if dbuser is not None and not dbuser.password_mustchange:
            # pylint: disable = no-member
            dbuser.set_session(request)
            log_info(request, 'autologin')
            return request.session['user']
        if dbuser is not None and dbuser.password_mustchange:
            request.session['user'] = {'login': ''}

        return None


# =============================================================================
class RootFactory(object):
    """Access Control List (ACL) definition.

    :type  request: pyramid.request.Request
    :param request:
        Current request.
    """
    # pylint: disable = too-few-public-methods
    __acl__ = [
        (Allow, Authenticated, 'authenticated'),
        (Allow, 'system.administrator', ALL_PERMISSIONS)] + [
            (Allow, '{0}.{1}'.format(group[0], principal[0]), principal[2])
            for group in PRINCIPALS for principal in group[2]]

    # -------------------------------------------------------------------------
    def __init__(self, request):
        """Constructor method."""



# -*- coding: utf-8 -*-
# pylint: disable = import-outside-toplevel
"""Tests of ``lib.xml`` classes and functions."""

from os.path import join, dirname
from io import BytesIO
from filecmp import cmpfiles
from shutil import copy
from collections import namedtuple
from json import dumps
from unittest import TestCase

from lxml import etree

from pyramid.i18n import TranslationString

from . import TmpDirTestCase


# =============================================================================
class ULibXmlLoadXml(TestCase):
    """Unit test class for :func:`lib.xml.load_xml`."""

    # -------------------------------------------------------------------------
    @classmethod
    def _call_fut(cls, filename, relaxngs='', data=None, xinclude=False):
        """Call function under test."""
        from ..lib.xml import load_xml

        if relaxngs == '':
            relaxngs = {
                'chrysalio': join(
                    dirname(__file__), '..', 'RelaxNG', 'chrysalio.rng')}
        return load_xml(filename, relaxngs, data, False, None, xinclude)

    # -------------------------------------------------------------------------
    def test_unknown_file(self):
        """[u:lib.xml.load_xml] unknown file"""
        tree = self._call_fut(join(dirname(__file__), 'Xml', 'test.usr.xml'))
        self.assertIsInstance(tree, (str, TranslationString))
        self.assertIn('Unknown file', tree)

    # -------------------------------------------------------------------------
    def test_not_well_formed(self):
        """[u:lib.xml.load_xml] not well-formed XML file"""
        from . import TEST1_USR_XML

        tree = self._call_fut(TEST1_USR_XML)
        self.assertIsInstance(tree, str)
        self.assertIn('Opening and ending tag mismatch', tree)

    # -------------------------------------------------------------------------
    def test_data_from_file(self):
        """[u:lib.xml.load_xml] XML data from file"""
        from . import TEST3_USR_XML

        with open(TEST3_USR_XML, 'r') as hdl:
            data = hdl.read()
        if not isinstance(data, bytes):
            data = bytes(data.encode())
        data = etree.parse(BytesIO(data))
        tree = self._call_fut(TEST3_USR_XML, data=data, xinclude=True)
        self.assertNotIsInstance(tree, (str, TranslationString))
        # pylint: disable = no-member
        self.assertEqual(tree.getroot().findtext('user/password'), 'user1pwd')

    # -------------------------------------------------------------------------
    def test_data_from_string(self):
        """[u:lib.xml.load_xml] XML data from string"""
        tree = self._call_fut(
            'test.xml', data='<chrysalio version="1.0"></chrysalio>')
        self.assertNotIsInstance(tree, (str, TranslationString))
        # pylint: disable = no-member
        self.assertEqual(tree.getroot().tag, 'chrysalio')

        tree = self._call_fut(
            'test.xml', data='<chrysalio version="1.0"></chrysalio>')
        self.assertNotIsInstance(tree, (str, TranslationString))
        self.assertEqual(tree.getroot().tag, 'chrysalio')

        tree = self._call_fut(
            'test.xml',
            data=bytes('<chrysalio version="1.0"></chrysalio>'.encode(
                'utf8')))
        self.assertNotIsInstance(tree, (str, TranslationString))
        self.assertEqual(tree.getroot().tag, 'chrysalio')

    # -------------------------------------------------------------------------
    def test_without_relaxng(self):
        """[u:lib.xml.load_xml] without Relax NG"""
        from . import TEST3_USR_XML

        tree = self._call_fut(TEST3_USR_XML, relaxngs=None)
        self.assertNotIsInstance(tree, str)

    # -------------------------------------------------------------------------
    def test_incorrect_xinclude(self):
        """[u:lib.xml.load_xml] incorrect xinclude"""

        data = \
            '<chrysalio version="1.0"'\
            '         xmlns:xi="http://www.w3.org/2001/XInclude">'\
            '  <xi:include href="chrysalio.inc.xml"/>'\
            '</chrysalio>'
        tree = self._call_fut('test.xml', data=data, xinclude=True)
        self.assertIsInstance(tree, str)
        self.assertIn('could not load chrysalio.inc.xml', tree)

    # -------------------------------------------------------------------------
    def test_invalid(self):
        """[u:lib.xml.load_xml] invalid XML file"""
        from . import TEST2_USR_XML

        tree = self._call_fut(TEST2_USR_XML)
        self.assertIsInstance(tree, (str, TranslationString))
        self.assertTrue('Line' in tree)


# =============================================================================
class ULibXmlValidateXml(TestCase):
    """Unit test class for :func:`lib.xml.validate_xml`."""

    # -------------------------------------------------------------------------
    @classmethod
    def _call_fut(cls, filename, relaxngs=None, noline=False):
        """Call function under test."""
        from ..lib.xml import validate_xml

        tree = etree.parse(filename)
        if not relaxngs:
            relaxngs = {
                'chrysalio':
                join(dirname(__file__), '..', 'RelaxNG', 'chrysalio.rng')}
        return validate_xml(tree, relaxngs, noline)

    # -------------------------------------------------------------------------
    def test_invalid(self):
        """[u:lib.xml.validate_xml] invalid XML file"""
        from . import TEST2_USR_XML

        error = self._call_fut(TEST2_USR_XML)
        self.assertIsInstance(error, (str, TranslationString))
        self.assertIn('Line', error)

    # -------------------------------------------------------------------------
    def test_valid(self):
        """[u:lib.xml.validate_xml] valid XML file"""
        from . import TEST3_USR_XML

        error = self._call_fut(TEST3_USR_XML)
        self.assertIsNone(error)

        error = self._call_fut(
            TEST3_USR_XML, relaxngs={
                'chrysalio, version=1.0':
                join(dirname(__file__), '..', 'RelaxNG', 'chrysalio.rng')})
        self.assertIsNone(error)

    # -------------------------------------------------------------------------
    def test_unknown_notfound(self):
        """[u:lib.xml.validate_xml] Relax NG not found"""
        from . import TEST3_USR_XML

        error = self._call_fut(
            TEST3_USR_XML, relaxngs={
                'chrysalio': join(dirname(__file__), 'Xml', 'chrysalio0.rng')})
        self.assertIsInstance(error, str)
        self.assertIn('failed to load', error)

    # -------------------------------------------------------------------------
    def test_unknown_relaxng(self):
        """[u:lib.xml.validate_xml] unknown Relax NG"""
        from . import TEST3_USR_XML

        error = self._call_fut(
            TEST3_USR_XML, relaxngs={
                'chrysalio0': join(dirname(__file__), 'Xml', 'chrysalio0.rng'),
                'chrysalio, version=0.1':
                join(dirname(__file__), 'Xml', 'chrysalio.rng')})
        self.assertIsInstance(error, (str, TranslationString))
        self.assertIn('Relax NG not found', error)

    # -------------------------------------------------------------------------
    def test_invalid_relaxng(self):
        """[u:lib.xml.validate_xml] invalid Relax NG"""
        from . import TEST3_USR_XML

        error = self._call_fut(
            TEST3_USR_XML, relaxngs={
                'chrysalio': join(dirname(__file__), 'Xml', 'chrysalio.rng')})
        self.assertIsInstance(error, str)
        self.assertIn('no matching definition', error)


# =============================================================================
class ULibXmlRelaxng4Validation(TestCase):
    """Unit test class for :func:`lib.xml.relaxng4validation`."""

    # -------------------------------------------------------------------------
    def test_it(self):
        """[u:lib.xml.relaxng4validation]"""
        from ..lib.xml import relaxng4validation

        # None as Relax NG
        relaxngs = relaxng4validation(None)
        self.assertIsNone(relaxngs)

        # No namespace, no version
        relaxngs = relaxng4validation({
            'root': 'article',
            'file': join(dirname(__file__), '..', 'RelaxNG', 'article.rng')})
        self.assertIn('article', relaxngs)

        # No namespace, version
        relaxngs = relaxng4validation({
            'root': 'chrysalio', 'version': '1.0',
            'file': join(dirname(__file__), '..', 'RelaxNG', 'chrysalio.rng')})
        self.assertIn('chrysalio, version=1.0', relaxngs)

        # Namespace, no version
        relaxngs = relaxng4validation({
            'namespace': 'http://ns.chrysal.io/cioskeleton',
            'root': 'cioskeleton',
            'file': join(
                dirname(__file__), '..', 'RelaxNG', 'cioskeleton.rng')})
        self.assertIn(
            '{http://ns.chrysal.io/cioskeleton}cioskeleton', relaxngs)

        # Namespace, version
        relaxngs = relaxng4validation({
            'namespace': 'http://ns.chrysal.io/cioprocessor',
            'root': 'cioprocessor', 'version': '1.0',
            'file': join(
                dirname(__file__), '..', 'RelaxNG', 'cioprocessor.rng')})
        self.assertIn(
            '{http://ns.chrysal.io/cioprocessor}cioprocessor, version=1.0',
            relaxngs)


# =============================================================================
class ULibXmlLoadXsl(TestCase):
    """Unit test class for :func:`lib.xml.load_xsl`."""

    # -------------------------------------------------------------------------
    @classmethod
    def _call_fut(cls, filename):
        """Call function under test."""
        from ..lib.xml import load_xslt
        return load_xslt(filename)

    # -------------------------------------------------------------------------
    def test_unknown_xsl(self):
        """[u:lib.xml.load_xsl] unknown XSL file"""
        xslt = self._call_fut(join(dirname(__file__), 'Xml', 'transform0.xsl'))
        self.assertIsInstance(xslt, str)
        self.assertIn('Error reading file', xslt)

    # -------------------------------------------------------------------------
    def test_valid_xsl(self):
        """[u:lib.xml.load_xsl] valid XSL file"""
        from . import TRANSFORM_XSL

        xslt = self._call_fut(TRANSFORM_XSL)
        self.assertNotIsInstance(xslt, str)


# =============================================================================
class ULibXmlCreateEntireXml(TestCase):
    """Unit test class for :func:`lib.xml.create_entire_xml`."""

    # -------------------------------------------------------------------------
    @classmethod
    def _call_fut(cls, elements):
        """Call function under test."""
        from ..lib.xml import create_entire_xml

        relaxng = {
            'root': 'chrysalio', 'version': '1.0',
            'file': join(dirname(__file__), '..', 'RelaxNG', 'chrysalio.rng')}

        return create_entire_xml(relaxng, elements)

    # -------------------------------------------------------------------------
    def test_invalid_element(self):
        """[u:lib.xml.create_entire_xml] invalid element"""
        user_elt = etree.XML(
            '<user>'
            '  <login>test2</login>'
            '  <password>test2pass</password>'
            '  <firstname>Sophie</firstname>'
            '  <lastname>FONFEC</lastname>'
            '  <email>test2_chrysal.io</email>'
            '</user>')

        error = self._call_fut([user_elt])
        # pylint: disable = protected-access
        self.assertNotIsInstance(error, etree._Element)
        # pylint: enable = protected-access
        self.assertIn('email failed to validate', error)

    # -------------------------------------------------------------------------
    def test_one_element(self):
        """[u:lib.xml.create_entire_xml] one element"""
        user_elt = etree.XML(
            '<user>'
            '  <login>test1</login>'
            '  <password>test1pass</password>'
            '  <firstname>Édith</firstname>'
            '  <lastname>AVULEUR</lastname>'
            '  <email>test1@chrysal.io</email>'
            '</user>')

        root_elt = self._call_fut([user_elt])
        # pylint: disable = protected-access
        self.assertIsInstance(root_elt, etree._Element)
        # pylint: enable = protected-access
        # pylint: disable = no-member
        self.assertEqual(root_elt.tag, 'chrysalio')
        self.assertEqual(root_elt.get('version'), '1.0')
        self.assertEqual(root_elt.findtext('user/login'), 'test1')
        self.assertEqual(root_elt.findtext('user/firstname'), 'Édith')

    # -------------------------------------------------------------------------
    def test_several_elements(self):
        """[u:lib.xml.create_entire_xml] several elements"""
        profile_elt = etree.XML(
            '<profile id="user_manager">'
            '  <label xml:lang="en">User manager</label>'
            '  <principals>'
            '    <principal>user.creator</principal>'
            '  </principals>'
            '</profile>')
        user1_elt = etree.XML(
            '<user>'
            '  <login>test1</login>'
            '  <password>test1pass</password>'
            '  <firstname>Édith</firstname>'
            '  <lastname>AVULEUR</lastname>'
            '  <email>test1@chrysal.io</email>'
            '</user>')
        user2_elt = etree.XML(
            '<user>'
            '  <login>test2</login>'
            '  <password>test2pass</password>'
            '  <firstname>Sophie</firstname>'
            '  <lastname>FONFEC</lastname>'
            '  <email>test2@chrysal.io</email>'
            '</user>')

        root_elt = self._call_fut([profile_elt, user1_elt, user2_elt])
        # pylint: disable = protected-access
        self.assertIsInstance(root_elt, etree._Element)
        # pylint: enable = protected-access
        # pylint: disable = no-member
        self.assertEqual(len(root_elt.findall('*')), 2)
        self.assertIsNotNone(root_elt.find('profiles'))
        self.assertIsNotNone(root_elt.find('users'))
        self.assertEqual(len(root_elt.findall('profiles/*')), 1)
        self.assertEqual(len(root_elt.findall('users/*')), 2)

    # -------------------------------------------------------------------------
    def test_settings_element(self):
        """[u:lib.xml.create_entire_xml] settings element"""
        settings_elt = etree.XML(
            '<settings>'
            '  <title>Chrysalio – Test</title>'
            '  <email>admin@chrysal.io</email>'
            '  <password-min-length>8</password-min-length>'
            '  <language>en</language>'
            '  <page-size>80</page-size>'
            '  <download-max-size>10485760</download-max-size>'
            '  <theme>Default</theme>'
            '</settings>')
        profile_elt = etree.XML(
            '<profile id="user_manager">'
            '  <label xml:lang="en">User manager</label>'
            '  <principals>'
            '    <principal>user.creator</principal>'
            '  </principals>'
            '</profile>')

        root_elt = self._call_fut([settings_elt, profile_elt])
        # pylint: disable = protected-access
        self.assertIsInstance(root_elt, etree._Element)
        # pylint: enable = protected-access
        # pylint: disable = no-member
        self.assertEqual(root_elt.tag, 'chrysalio')
        self.assertEqual(root_elt.get('version'), '1.0')
        self.assertEqual(root_elt.findtext(
            'settings/title'), 'Chrysalio – Test')

    # -------------------------------------------------------------------------
    def test_module(self):
        """[u:lib.xml.create_entire_xml] with a module"""
        skeleton_elt = etree.XML(
            '<module id="chrysalio.modules.cioskeleton" inactive="true">'
            '  <cioskeleton xmlns="http://ns.chrysal.io/chrysalio/skeleton"'
            '              version="1.0">'
            '    <bones>'
            '      <bone id="foot">'
            '        <label>Foot</label>'
            '      </bone>'
            '    </bones>'
            '  </cioskeleton>'
            '</module>')
        root = self._call_fut([skeleton_elt])
        self.assertTrue(root.xpath('module'))
        self.assertTrue(
            root.xpath('module[@id="chrysalio.modules.cioskeleton"]'))

        ldap_elt = etree.XML(
            '<module id="chrysalio.modules.ldap" inactive="true"/>')
        root_elt = self._call_fut([ldap_elt, skeleton_elt])
        self.assertTrue(root_elt.xpath('modules'))
        self.assertTrue(root_elt.xpath(
            'modules/module[@id="chrysalio.modules.ldap"]'))
        self.assertTrue(root_elt.xpath(
            'modules/module[@id="chrysalio.modules.cioskeleton"]'))

    # -------------------------------------------------------------------------
    def test_namespace_element(self):
        """[u:lib.xml.create_entire_xml] with namespace"""
        from ..lib.xml import create_entire_xml

        namespace = 'http://ns.chrysal.io/cioskeleton'
        relaxng = {
            'namespace': namespace, 'root': 'cioskeleton', 'version': '1.0',
            'file': join(
                dirname(__file__), '..', 'modules', 'cioskeleton', 'RelaxNG',
                'cioskeleton.rng')}
        skeleton_elt = etree.XML(
            '<bone id="hand">'
            '  <label>Hand</label>'
            '  <attachments key="Hand">'
            '    <picture>hand.png</picture>'
            '  </attachments>'
            '</bone>')

        root_elt = create_entire_xml(relaxng, (skeleton_elt,))
        # pylint: disable = protected-access
        self.assertIsInstance(root_elt, etree._Element)
        # pylint: enable = protected-access
        # pylint: disable = no-member
        self.assertEqual(root_elt.tag, '{{{0}}}cioskeleton'.format(namespace))
        self.assertEqual(root_elt.get('version'), '1.0')
        self.assertTrue(root_elt.xpath(
            'ns0:bones/ns0:bone[@id="hand"]', namespaces={'ns0': namespace}))


# =============================================================================
class ULibXmlI18nXmlText(TestCase):
    """Unit test class for :func:`lib.xml.i18n_xml_text`."""

    # -------------------------------------------------------------------------
    @classmethod
    def _call_fut(cls, root_elt=None, xpath='label'):
        """Call function under test."""
        from . import I18N_XML
        from ..lib.xml import i18n_xml_text

        if root_elt is None:
            tree = etree.parse(I18N_XML)
            root_elt = tree.getroot().find('profile')
        return i18n_xml_text(root_elt, xpath)

    # -------------------------------------------------------------------------
    def test_it(self):
        """[u:lib.xml.i18n_xml_text]"""
        labels = self._call_fut()
        self.assertIsInstance(labels, dict)
        self.assertIn('en', labels)
        self.assertIn('fr', labels)
        self.assertEqual(labels['en'], 'Profile Manager')


# =============================================================================
class ULibXmlDb2xmlI18nLabels(TestCase):
    """Unit test class for :func:`lib.xml.db2xml_i18n_labels`."""

    # -------------------------------------------------------------------------
    def test_it(self):
        """[u:lib.xml.db2xml_i18n_labels]"""
        from ..lib.xml import db2xml_i18n_labels

        dbitem = namedtuple(
            'DBItem', 'i18n_label i18n_description')(
                i18n_label=dumps({
                    'en': 'Item label', 'fr': "Libellé de l'élément"}),
                i18n_description={'en': 'Item description'})
        root_elt = etree.Element('root')

        db2xml_i18n_labels(dbitem, root_elt, 1)
        self.assertEqual(len(root_elt.xpath('label')), 2)
        self.assertEqual(len(root_elt.xpath('description')), 1)


# =============================================================================
class ULibXmlXmlWrap(TestCase):
    """Unit test class for :func:`lib.xml.xml_wrap`."""

    # -------------------------------------------------------------------------
    def test_it(self):
        """[u:lib.xml.xml_wrap]"""
        from ..lib.xml import xml_wrap

        result = xml_wrap(
            'Lorèm ipsum dolor sit amet, consectetur adipiscing elit. '
            'Sed in ultricies ex, ut sollicitudin velit. Praesent semper '
            'venenatis pellentesque.', 2)
        self.assertEqual(
            result,
            '\n      Lorèm ipsum dolor sit amet, consectetur adipiscing '
            'elit. Sed in ultricies\n'
            '      ex, ut sollicitudin velit. Praesent semper venenatis '
            'pellentesque.\n    ')


# =============================================================================
class ULibXmlCheckChrysalioRng(TmpDirTestCase):
    """Unit test class for :func:`lib.xml.check_chrysalio_rng`."""

    # -------------------------------------------------------------------------
    def test_it(self):
        """[u:lib.xml.check_chrysalio_rng]"""
        from ..lib.xml import check_chrysalio_rng
        from . import TEST_DIR

        relaxng_dir = TEST_DIR
        cio_relaxng_dir = join(dirname(__file__), '..', 'RelaxNG')
        copy(join(cio_relaxng_dir, 'chrysalio.rnc'),
             join(relaxng_dir, 'chrysalio.rnc'))
        with open(join(relaxng_dir, 'chrysalio.rng'), 'w'):
            pass
        compare = cmpfiles(
            cio_relaxng_dir, relaxng_dir, ('chrysalio.rnc', 'chrysalio.rng'))
        self.assertEqual(compare[0], ['chrysalio.rnc'])
        self.assertEqual(compare[1], ['chrysalio.rng'])
        self.assertEqual(len(compare[2]), 0)

        check_chrysalio_rng(relaxng_dir)

        compare = cmpfiles(
            cio_relaxng_dir, relaxng_dir, ('chrysalio.rnc', 'chrysalio.rng'))
        self.assertEqual(len(compare[0]), 2)
        self.assertEqual(len(compare[1]), 0)
        self.assertEqual(len(compare[2]), 0)



# -*- coding: utf-8 -*-
"""Class to manage a filter to select items among a collection."""

from collections import OrderedDict
from json import loads
from re import sub

from webhelpers2.html import literal
from sqlalchemy.sql import text

from pyramid.httpexceptions import HTTPForbidden
from .i18n import _, translate_field
from .utils import shorten


OPERATORS = OrderedDict((('AND', _('AND')), ('OR', _('OR'))))
COMPARISONS = OrderedDict((
    ('=', '='), ('!=', '≠'), ('>', '>'), ('>=', _('≥')), ('<', '<'),
    ('<=', _('≤'))))


# =============================================================================
class Filter(object):
    """A class to manage filters.

    :type  request: pyramid.request.Request
    :param request:
        Current request.
    :param str filter_id:
        Filter ID.
    :param list inputs:
        A list of inputs like
        ``((key, label, is_static, available_values),...)``.
    :param list initials: (optional)
        Initial conditions.
    :param str remove: (optional)
        Number of the condition to remove from the previous filter.
    :type  comparisons: collections.OrderedDict
    :param comparisons: (optional)
        Customize list of comparaisons.

    For ``inputs`` the ``available_values`` field must be:

    * ``None`` for a string value
    * ``''`` for a string value with auto-completion
    * ``True`` for a boolean value
    * ``0`` for an integer value
    * a ``list`` of value/label for closed list of values.

    A filter is a list of AND-conditions.
    Each AND-condition is a list of OR-conditions.
    An OR-condition is a tuple such as ``(key, comparison, value)``.

    The ``comparison`` is one of the items of ``comparisons`` and can be
    ``'='``, ``'!='``, ``'>'``, ``'>='``, ``'<'`` or ``'<='``.
    The ``value`` can be a boolean, a string or a
    :class:`~pyramid.i18n.TranslationString` instance.

    For instance, the filter:

    ``(id=2 OR id>5) AND active AND group!='Foo'`` is stored as:

    ``[[('id', '=', 2), ('id', '>', 5)], [('active', '=', True)],
    [('group', '!=', 'Foo')]]``
    """

    # -------------------------------------------------------------------------
    def __init__(
            self, request, filter_id, inputs, initials=None, remove=None,
            comparisons=None):
        """Constructor method."""
        # pylint: disable = too-many-arguments
        self.uid = filter_id
        self._request = request
        self._inputs = OrderedDict(
            [(k[0], (k[1], k[2], k[3])) for k in inputs])
        self._comparisons = comparisons \
            if comparisons is not None else COMPARISONS

        # Retrieve last filter
        if 'filters' not in request.session:
            request.session['filters'] = {}
        self._conditions = request.session['filters'].get(filter_id) or []

        # Append initial conditions
        if filter_id not in request.session['filters'] \
           and initials is not None:
            for condition in initials:
                self.append_condition(*condition)

        # Remove obsolete conditions
        if remove is not None:
            self.remove_condition(remove)

        # Append new conditions
        if 'filter' in request.POST:
            for key in self._inputs:
                if self._inputs[key][1] and \
                   request.POST.get('filter_value_{0}'.format(key)):
                    self.append_condition(
                        key, request.POST.get(
                            'filter_comparison_{0}'.format(key), '='),
                        request.POST['filter_value_{0}'.format(key)],
                        request.POST.get(
                            'filter_operator_{0}'.format(key), 'AND'))
            self.append_condition(
                request.POST.get('filter_key'),
                request.POST.get('filter_comparison', '='),
                request.POST.get('filter_value'),
                request.POST.get('filter_operator', 'AND'))

    # -------------------------------------------------------------------------
    def __str__(self):
        """String representation."""
        return str(self._conditions)

    # -------------------------------------------------------------------------
    def is_empty(self):
        """Return ``True`` if a filter is empty."""
        return not bool(self._conditions)

    # -------------------------------------------------------------------------
    def clear(self):
        """Clear filter."""
        self._conditions = []
        self._request.session['filters'][self.uid] = []

    # -------------------------------------------------------------------------
    def append_condition(self, key, comparison, value, operator='AND'):
        """Append an AND-condition to the filter.

        :param str key:
            The key to use.
        :param str comparison: (``'='``, ``'!='``, ``'>'``, ``'<'``,...)
            How to compare.
        :type  value: :class:`str`, :class:`bool` or
            :class:`~pyramid.i18n.TranslationString`
        :param value:
            The value of the filter. It is optional for boolean filter and its
            default value is ``True``.
        :param str operator: ('AND' or 'OR', default='AND')
            The operator to use to complete the filter.
        """
        if not key or not value or key not in self._inputs or \
           comparison not in self._comparisons or operator not in OPERATORS:
            return
        if self._inputs[key][2] is True:
            value = True
        elif self._inputs[key][2] == 0:
            value = int(sub(r'[^\d-]+', '', value) or '0')
        condition = (key, comparison, value)
        if self._conditions and operator == 'OR':
            if condition not in self._conditions[-1]:
                self._conditions[-1].append(condition)
        elif [condition] not in self._conditions:
            self._conditions.append([condition])
        self._request.session['filters'][self.uid] = self._conditions

    # -------------------------------------------------------------------------
    def remove_condition(self, index):
        """Remove one AND-condition of the filter.

        :param str index:
            Index of the AND-condition to remove.
        """
        if index is None or not index.isdigit():
            return
        index = int(index)
        if index < 0 or index >= len(self._conditions):
            return
        del self._conditions[index]
        self._request.session['filters'][self.uid] = self._conditions

    # -------------------------------------------------------------------------
    def html_conditions(self):
        """An HTML representation of the current conditions.

        :rtype: webhelpers2.html.literal
        """
        if not self._conditions:
            return ''

        html = ''
        for and_num, and_condition in enumerate(self._conditions):
            if and_num:
                html += ' <strong>{0}</strong> '.format(
                    self._request.localizer.translate(OPERATORS['AND']))
            html += '<span class="cioFilterCondition">'\
                '<input type="submit" value="X" name="crm!{0}.x"/>'.format(
                    and_num)
            for or_num, or_condition in enumerate(and_condition):
                if or_num:
                    html += ' <strong>{0}</strong> '.format(
                        self._request.localizer.translate(OPERATORS['OR']))
                condition = self._inputs[or_condition[0]]
                html += '<span>{0}{1}'.format(
                    condition[2] is True and or_condition[1] == '!=' and
                    self._request.localizer.translate(_('EXCEPT')) + ' ' or '',
                    self._request.localizer.translate(condition[0]))
                if isinstance(condition[2], (list, tuple)):
                    html += ' {0} {1}'.format(
                        self._comparisons[or_condition[1]],
                        self._request.localizer.translate(
                            dict(condition[2]).get(
                                or_condition[2], or_condition[2])))
                elif condition[2] is not True:
                    html += ' {0} {1}'.format(
                        self._comparisons[or_condition[1]], or_condition[2])
                html += '</span>'
            html += '</span> '
        return literal(html)

    # -------------------------------------------------------------------------
    def html_inputs(self, form, tag='span'):
        """An HTML representation of the current inputs.

        :type  form: .lib.form.Form
        :param form:
            Current form.
        :param str tag: (Default='div')
            Tag which wraps each input line.
        :rtype: webhelpers2.html.literal
        """
        # Static inputs
        html = ''
        has_static = False
        inputs = []
        translate = self._request.localizer.translate
        for key in self._inputs:
            if self._conditions or not self._inputs[key][1]:
                inputs.append((key, self._inputs[key][0]))
                continue
            has_static = True
            html += literal(
                '<{0} class="cioFilterInput cioFilterInputStatic">'
                '<span>'.format(tag))
            html += literal('<span class="cioFilterNoOp"> </span>')
            html += literal(
                '<span class="cioFilterKey" title="{0}">'
                '<span>{1}</span></span>'.format(
                    translate(_('Filter name')),
                    translate(self._inputs[key][0])))
            html += self._html_select_comparison(form, key)
            html += self._html_select_value(form, key)
            html += literal('</span></{0}>'.format(tag))
        if not inputs:
            return literal(html)

        # Dynamic inputs
        html += literal('<{0} class="cioFilterInput"><span>'.format(tag))
        key = self._request.POST.get('filter_key') or \
            not has_static and inputs[0][0]
        if self._conditions:
            html += literal('<span class="cioFilterOp">') + \
                form.select('filter_operator', None, OPERATORS.items()) + \
                literal('</span>')
        else:
            html += literal('<span class="cioFilterNoOp"> </span>')
        if has_static:
            inputs.insert(0, ('', ' '))
        html += literal('<span class="cioFilterKey" title="{0}">'.format(
            translate(_('Filter name')))) + form.select(
                'filter_key', None, inputs, True) + literal('</span>')
        html += self._html_select_comparison(form, key, 'filter_comparison')
        html += self._html_select_value(form, key, 'filter_value')
        html += literal('</span></{0}>'.format(tag))

        return html

    # -------------------------------------------------------------------------
    def sql(self, dbquery, table_name, ignored=None):
        """Complete a SqlAlchemy query with of the current filter.

        :param sqlalchemy.orm.query.Query dbquery:
            SqlAlchemy query to complete.
        :param str table_name:
            Name of the default SQL table.
        :param list ignored: (optional)
            List of and-conditions that must be ignored. They should be
            processed by the caller, probably because of joins.
        :rtype: sqlalchemy.orm.query.Query
        """
        and_conditions = self._sql_remove_ignored(ignored)
        if not and_conditions:
            return dbquery

        like = 'ILIKE' if dbquery.session.bind.dialect.name == 'postgresql' \
            else 'LIKE'
        for and_num, and_condition in enumerate(and_conditions):
            clause = '('
            params = {}
            for or_num, or_condition in enumerate(and_condition):
                if or_num:
                    clause += ' OR '
                available_values = self._inputs[or_condition[0]][2]
                if or_condition[1] == '=' and available_values in (None, ''):
                    comparison = like
                elif or_condition[1] == '!=' and available_values in (
                        None, ''):
                    comparison = 'NOT {0}'.format(like)
                else:
                    comparison = or_condition[1]
                value = True if available_values is True else or_condition[2]
                clause += '{0}.{1} {2} :p{3}{4:02d}'.format(
                    table_name, or_condition[0], comparison, and_num, or_num)
                params['p{0}{1:02d}'.format(and_num, or_num)] = \
                    '%{0}%'.format(value) if like in comparison else value
            clause += ')'
            dbquery = dbquery.filter(text(clause).bindparams(**params))

        return dbquery

    # -------------------------------------------------------------------------
    @classmethod
    def sql_autocomplete(
            cls, request, table_class, where=None, limit=10, width=0):
        """suggest terms for auto-completion in SQL

        :type  request: pyramid.request.Request
        :param request:
            Current request.
        :param table_class:
            SQLAlchemy table class to query.
        :param dict where: (optional)
            WHERE clause for the query.
        :param int limit: (default=10)
            Limit for the search.
        :param int width: (default=25)
            Shorten the length of each suggestion.
        :rtype: list
        """
        field = request.params.get('field')
        if not request.is_xhr or not hasattr(table_class, field):
            raise HTTPForbidden()

        value = '%{0}%'.format(request.params.get('value'))
        dbquery = request.dbsession.query(getattr(table_class, field)).filter(
            getattr(table_class, field).ilike(value))
        if where is not None:
            dbquery = dbquery.filter_by(**where)
        data = [k[0] for k in dbquery.group_by(field).limit(limit)]

        if field == 'i18n_label':
            data = [translate_field(request, loads(k)) for k in data]
        if width:
            data = [shorten(k, width, '') for k in data]
        return data

    # -------------------------------------------------------------------------
    def whoosh(self):
        """Return the current filter in Whoosh query language.

        :rtype: tuple
        :return:
            a tuple search as ``(fieldnames, query)``.
        """
        # pylint: disable = too-many-branches
        fieldnames = set()
        clause = ''
        for and_num, and_condition in enumerate(self._conditions):
            if and_num:
                clause += ' AND '
            clause += '('
            for or_num, or_condition in enumerate(and_condition):
                if or_num:
                    clause += ' OR '
                fieldnames.add(or_condition[0])
                comparison = or_condition[1]
                if or_condition[2] is True:
                    clause += '{0}{1}:TRUE'.format(
                        comparison == '!=' and 'NOT ' or '', or_condition[0])
                elif comparison == '=' and not isinstance(
                        or_condition[2], (int, float, bool)) and \
                        ' ' in or_condition[2]:
                    clause += '{0}:"{1}"'.format(
                        or_condition[0], or_condition[2])
                elif comparison == '=':
                    clause += '{0}:({1})'.format(
                        or_condition[0], or_condition[2])
                elif comparison == '!=' and not isinstance(
                        or_condition[2], (int, float, bool)) \
                        and ' ' in or_condition[2]:
                    clause += 'NOT {0}:"{1}"'.format(
                        or_condition[0], or_condition[2])
                elif comparison == '!=':
                    clause += 'NOT {0}:({1})'.format(
                        or_condition[0], or_condition[2])
                elif comparison == '>':
                    clause += '{0}:{{{1} TO}}'.format(
                        or_condition[0], or_condition[2])
                elif comparison == '>=':
                    clause += '{0}:[{1} TO]'.format(
                        or_condition[0], or_condition[2])
                elif comparison == '<':
                    clause += '{0}:{{TO {1}}}'.format(
                        or_condition[0], or_condition[2])
                elif comparison == '<=':
                    clause += '{0}:[TO {1}]'.format(
                        or_condition[0], or_condition[2])
            clause += ')'

        return tuple(fieldnames), clause

    # -------------------------------------------------------------------------
    def _html_select_comparison(self, form, key, name=None):
        """HTML form select to choose the comparison operator of a condition.

        :type  form: .lib.form.Form
        :param form:
            Current form.
        :param str key:
            Current condition key.
        :param str name: (optional)
            Name of the form select.
        :rtype: webhelpers2.html.literal
        """
        html = literal('<span class="cioFilterComparison" title="{0}">'.format(
            self._request.localizer.translate(_('Comparison'))))
        if name is None:
            name = 'filter_comparison_{0}'.format(key)
        if key and self._inputs[key][2] not in (None, 0):
            return html + form.select(
                name, None, (('=', '='), ('!=', '≠'))) + literal('</span>')
        return html + form.select(name, None, self._comparisons.items()) + \
            literal('</span>')

    # -------------------------------------------------------------------------
    def _html_select_value(self, form, key, name=None):
        """HTML form select to choose the value of a condition.

        :type  form: .lib.form.Form
        :param form:
            Current form.
        :param str key:
            Current condition key.
        :param str name: (optional)
            Name of the form select.
        :rtype: webhelpers2.html.literal
        """
        html = literal('<span class="cioFilterValue" title="{0}">'.format(
            self._request.localizer.translate(_('Filter value'))))
        if name is None:
            name = 'filter_value_{0}'.format(key)
        if key and self._inputs[key][2] is True:
            return html + form.custom_checkbox(name) + literal('</span>')
        if key and self._inputs[key][2] not in (None, '', 0):
            return html + form.select(
                name, None, self._inputs[key][2]) + literal('</span>')
        if key and self._inputs[key][2] == '':
            return html + form.text(
                name, placeholder=self._request.localizer.translate(
                    _('filter here...')), class_='cioAutocomplete') + literal(
                        '</span>')
        return html + form.text(
            name, placeholder=self._request.localizer.translate(
                _('filter here...'))) + literal('</span>')

    # -------------------------------------------------------------------------
    def _sql_remove_ignored(self, ignored):
        """Remove ignored conditions.

        :param list ignored:
            List of and-conditions that must be ignored. They should be
            processed by the caller, probably because of joins.
        :rtype: list
        """
        if ignored is None:
            return self._conditions

        and_conditions = []
        for and_condition in self._conditions:
            new_and_conditions = []
            for or_condition in and_condition:
                if or_condition[0] not in ignored:
                    new_and_conditions.append(or_condition)
            if new_and_conditions:
                and_conditions.append(new_and_conditions)

        return and_conditions



# -*- coding: utf-8 -*-
# pylint: disable = import-outside-toplevel
"""Tests of ``lib.panel`` class."""

from collections import namedtuple

from pyramid.testing import DummySession

from . import DBUnitTestCase
from ..lib.panel import Panel


# =============================================================================
class DummyPanel(Panel):
    """A class to test panels."""

    uid = 'dummy'
    label = 'Side Panel'


# =============================================================================
class OtherDummyPanel(Panel):
    """Another class to test panels."""

    uid = 'other'
    label = 'Other Side Panel'


# =============================================================================
class ULibPanellPanel(DBUnitTestCase):
    """Unit test class for :class:`lib.panel.Panel`."""

    # -------------------------------------------------------------------------
    def test_register(self):
        """[u:lib.panel.Panel.register]"""
        # Register a non derived Panel
        panel = Panel.register(
            self.request.registry, Panel, area=('home',), add2systray=True)
        self.assertIsNone(panel)

        # Register
        panel = Panel.register(
            self.request.registry, DummyPanel, area=('home',),
            add2systray=True)
        self.assertIn('panels', self.request.registry)
        self.assertIn(DummyPanel.uid, self.request.registry['panels'])
        self.assertEqual(
            self.request.registry['panels'][DummyPanel.uid], panel)
        self.assertIsInstance(panel, Panel)
        self.assertEqual(panel.uid, DummyPanel.uid)
        self.assertIsInstance(panel.area, tuple)
        self.assertEqual(panel.area[0], 'home')
        self.assertIn('systray', self.request.registry)
        self.assertEqual(len(self.request.registry['systray']), 1)
        self.assertEqual(
            self.request.registry['systray'][0].uid, panel.uid)

        # Register an already registred panel
        Panel.register(self.request.registry, DummyPanel)
        panel = self.request.registry['panels'][DummyPanel.uid]
        self.assertEqual(panel.uid, DummyPanel.uid)

    # -------------------------------------------------------------------------
    def test_has_open_panel(self):
        """[u:lib.panel.Panel.has_open_panel]"""
        # Without panels
        self.assertIsNone(Panel.has_open_panel(self.request))

        # With a closed panel
        panel = Panel.register(self.request.registry, DummyPanel)
        self.assertFalse(panel.is_open(self.request))
        self.assertIsNone(Panel.has_open_panel(self.request))

        # With an open panel
        panel.open(self.request)
        self.assertIsNotNone(Panel.has_open_panel(self.request))

    # -------------------------------------------------------------------------
    def test_is_open(self):
        """[u:lib.panel.Panel.is_open]"""
        # In a right area
        self.request.matched_route = namedtuple('Route', 'name')(name='home')
        panel = Panel.register(
            self.request.registry, DummyPanel, area=('home',))
        self.assertFalse(panel.is_open(self.request))
        panel.open(self.request)
        self.assertTrue(panel.is_open(self.request))

        # Out of a right area
        self.request.matched_route = namedtuple('Route', 'name')(
            name='user_index')
        self.assertFalse(panel.is_open(self.request))

    # -------------------------------------------------------------------------
    def test_was_open(self):
        """[u:lib.panel.Panel.was_open]"""
        panel = Panel.register(self.request.registry, DummyPanel)
        self.assertFalse(panel.was_open(self.request))
        panel.open(self.request)
        self.assertFalse(panel.was_open(self.request))
        panel.open(self.request)
        self.assertTrue(panel.was_open(self.request))

    # -------------------------------------------------------------------------
    def test_open(self):
        """[u:lib.panel.Panel.open]"""
        panel = Panel.register(self.request.registry, DummyPanel)
        panel_other = Panel.register(self.request.registry, OtherDummyPanel)
        panel_other.open(self.request)
        panel.open(self.request)
        self.assertTrue(panel.is_open(self.request))
        self.assertFalse(panel_other.is_open(self.request))

    # -------------------------------------------------------------------------
    def test_close(self):
        """[u:lib.panel.Panel.close]"""
        panel = Panel.register(self.request.registry, DummyPanel)
        panel.open(self.request)
        panel.close(self.request)
        self.assertFalse(panel.was_open(self.request))
        self.assertFalse(panel.is_open(self.request))

    # -------------------------------------------------------------------------
    def test_clear_values(self):
        """[u:lib.panel.Panel.clear_values]"""
        panel = DummyPanel()
        self.request.session = DummySession({'panels': {panel.uid: {
            'values': {'foo': 'bar'}}}})
        panel.clear_values(self.request)
        self.assertNotIn('values', self.request.session['panels'][panel.uid])

    # -------------------------------------------------------------------------
    def test_set_values(self):
        """[u:lib.panel.Panel.set_values]"""
        panel = DummyPanel()
        panel.set_values(self.request, {'foo': 'bar'})
        self.assertEqual(panel.value(self.request, 'foo'), 'bar')

    # -------------------------------------------------------------------------
    def test_values(self):
        """[u:lib.panel.Panel.values]"""
        self.request.session = DummySession({'panels': {DummyPanel.uid: {
            'values': {'foo': 'bar'}}}})
        values = DummyPanel().values(self.request)
        self.assertIsInstance(values, dict)
        self.assertIn('foo', values)
        self.assertEqual(values['foo'], 'bar')

    # -------------------------------------------------------------------------
    def test_set_value(self):
        """[u:lib.panel.Panel.set_value]"""
        panel = DummyPanel()
        panel.set_value(self.request, 'foo', 'bar')
        self.assertEqual(panel.value(self.request, 'foo'), 'bar')

    # -------------------------------------------------------------------------
    def test_value(self):
        """[u:lib.panel.Panel.value]"""
        panel = DummyPanel()
        self.request.session = DummySession({'panels': {panel.uid: {
            'values': {'huge': 'énorme'}}}})
        self.assertEqual(panel.value(self.request, 'huge'), 'énorme')
        self.assertIsNone(panel.value(self.request, 'none'))

    # -------------------------------------------------------------------------
    def test_render(self):
        """[u:lib.panel.Panel.render]"""
        self.configurator.add_route('home', '/')
        self.request.matched_route = namedtuple('Route', 'name')(name='home')
        panel = Panel.register(self.request.registry, DummyPanel)
        panel.need_form = True
        html = panel.render(self.request)
        self.assertIn('data-uid="{0}"'.format(panel.uid), html)
        self.assertIn('href="/?panel={0}"'.format(panel.uid), html)

    # -------------------------------------------------------------------------
    def test_route(self):
        """[u:lib.panel.Panel.route]"""
        panel = DummyPanel()
        self.configurator.add_route('home', '/')
        self.request.GET = {'panel': panel.uid}

        route = panel.route(self.request)
        self.assertEqual(route, '')

        self.request.matched_route = namedtuple('Route', 'name')(name='home')
        route = panel.route(self.request)
        self.assertEqual(route, '/?panel={0}'.format(panel.uid))

    # -------------------------------------------------------------------------
    def test_open_panel_css(self):
        """[u:lib.panel.Panel.open_panel_css]"""
        # Without panel
        css = Panel.open_panel_css(self.request)
        self.assertIsInstance(css, tuple)
        self.assertEqual(css, ())

        # Without open panel
        panel = Panel.register(self.request.registry, DummyPanel)
        panel.css = ('/cioxml/css/panel_cioset.css',)
        self.assertFalse(Panel.open_panel_css(self.request))

        # With open panel
        panel.open(self.request)
        css = Panel.open_panel_css(self.request)
        self.assertEqual(len(css), 1)
        self.assertEqual(css[0], '/cioxml/css/panel_cioset.css')

    # -------------------------------------------------------------------------
    def test_open_panel_js(self):
        """[u:lib.panel.Panel.open_panel_js]"""
        # Without panel
        javascripts = Panel.open_panel_js(self.request)
        self.assertIsInstance(javascripts, tuple)
        self.assertEqual(javascripts, ())

        # Without open panel
        panel = Panel.register(self.request.registry, DummyPanel)
        panel.javascripts = ('/cioxml/js/panel_cioset.js',)
        self.assertFalse(Panel.open_panel_js(self.request))

        # With open panel
        panel.open(self.request)
        javascripts = Panel.open_panel_js(self.request)
        self.assertEqual(len(javascripts), 1)
        self.assertEqual(javascripts[0], '/cioxml/js/panel_cioset.js')

    # -------------------------------------------------------------------------
    def test_manage_panels(self):
        """[u:lib.panel.Panel.manage_panels]"""
        # Without panel
        Panel.manage_panels(self.request)
        self.assertNotIn('panels', self.request.registry)

        # Open a panel
        panel = Panel.register(self.request.registry, DummyPanel)
        other_panel = Panel.register(self.request.registry, OtherDummyPanel)
        other_panel.open(self.request)
        self.assertTrue(other_panel.is_open(self.request))
        self.request.GET = {'panel': panel.uid}
        Panel.manage_panels(self.request)
        self.assertTrue(panel.is_open(self.request))
        self.assertFalse(panel.was_open(self.request))
        self.assertFalse(other_panel.is_open(self.request))

        # With panels, close
        Panel.manage_panels(self.request)
        self.assertFalse(panel.is_open(self.request))
        self.assertFalse(panel.was_open(self.request))



# -*- coding: utf-8 -*-
"""SQLAlchemy-powered model definitions for LDAP."""

from sqlalchemy import Column, Integer, String, Boolean, ForeignKey
from sqlalchemy.orm import relationship
from lxml import etree
from webhelpers2.html import HTML, literal
import colander

from ....lib.i18n import _
from ....lib.utils import encrypt, make_id, deltatime_label
from ....models import DBDeclarativeClass, ID_LEN
from ..relaxng import RELAXNG_CIOLDAP

LDAP_SUFFIX = 'cioldap'
LDAP_USER_FILTER = '(&(objectclass=inetOrgPerson)(uid=_UID_))'
LDAP_FIELD_FIRSTNAME = 'givenName'
LDAP_FIELD_LASTNAME = 'sn'
LDAP_FIELD_EMAIL = 'mail'
LDAP_CHECK_INTERVAL = 3600


# =============================================================================
class DBLdap(DBDeclarativeClass):
    """SQLAlchemy-powered LDAP class."""

    __tablename__ = 'ldap'
    __table_args__ = {'mysql_engine': 'InnoDB'}

    ldap_id = Column(Integer, primary_key=True)
    host = Column(String(128), default='localhost')
    port = Column(Integer, default=389)
    ssl = Column(Boolean(name='ssl'), default=False)
    check_interval = Column(Integer, default=LDAP_CHECK_INTERVAL)
    root_dn = Column(String(128))
    root_password = Column(String(128))
    base = Column(String(128), nullable=False)
    user_dn = Column(String(128), nullable=False)
    user_filter = Column(String(128), nullable=False)
    field_first_name = Column(String(32))
    field_last_name = Column(String(32), default='sn')
    field_email = Column(String(32), default='mail')

    user_profiles = relationship('DBLdapProfile', cascade='all, delete')

    suffix = LDAP_SUFFIX

    # -------------------------------------------------------------------------
    @classmethod
    def xml2db(cls, dbsession, ldap_elt):
        """Load LDAP settings from a XML element.

        :type  dbsession: sqlalchemy.orm.session.Session
        :param dbsession:
            SQLAlchemy session.
        :type  ldap_elt: lxml.etree.Element
        :param ldap_elt:
            Ldap XML element.
        :rtype: :class:`pyramid.i18n.TranslationString` or ``None``
        :return:
            Error message or ``None``.
        """
        if ldap_elt is None:
            return None

        # Clean up old settings
        for dbldap in dbsession.query(cls):
            dbsession.delete(dbldap)

        # Create LDAP settings
        ldap_record = cls.record_from_xml(ldap_elt)
        error = cls.record_format(ldap_record)
        if error:
            return error
        dbldap = cls(**ldap_record)
        dbsession.add(dbldap)

        # Add profiles
        dbsession.flush()
        namespace = '{{{0}}}'.format(RELAXNG_CIOLDAP['namespace'])
        user_elt = ldap_elt.find('{0}user'.format(namespace))
        for elt in user_elt.xpath('ns0:profiles/ns0:profile', namespaces={
                'ns0': RELAXNG_CIOLDAP['namespace']}):
            dbldap.user_profiles.append(
                DBLdapProfile(profile_id=make_id(elt.text, 'token', ID_LEN)))
        return None

    # -------------------------------------------------------------------------
    @classmethod
    def record_from_xml(cls, ldap_elt):
        """Convert a LDAP XML element into a dictionary.

        :type  ldap_elt: lxml.etree.Element
        :param ldap_elt:
            LDAP XML element.
        :rtype: dict
        """
        namespace = '{{{0}}}'.format(RELAXNG_CIOLDAP['namespace'])
        user_elt = ldap_elt.find('{0}user'.format(namespace))

        return {
            'host': ldap_elt.findtext(
                '{0}host'.format(namespace)) or 'localhost',
            'port': int(
                ldap_elt.findtext('{0}port'.format(namespace)) or '389'),
            'ssl': ldap_elt.findtext('{0}ssl'.format(namespace)) == 'true',
            'check_interval': int(
                ldap_elt.findtext('{0}check-interval'.format(namespace)) or
                LDAP_CHECK_INTERVAL),
            'root_dn': ldap_elt.findtext('{0}dn'.format(namespace)),
            'root_password': ldap_elt.findtext(
                '{0}password'.format(namespace)),
            'base': ldap_elt.findtext('{0}base'.format(namespace)),
            'user_dn': user_elt.findtext('{0}dn'.format(namespace)),
            'user_filter': (
                user_elt.findtext('{0}filter'.format(namespace)) or
                '(&(objectclass=inetOrgPerson)(uid=_UID_))'),
            'field_first_name': user_elt.findtext(
                '{0}first-name'.format(namespace)),
            'field_last_name': user_elt.findtext(
                '{0}last-name'.format(namespace)) or 'sn',
            'field_email': user_elt.findtext(
                '{0}email'.format(namespace)) or 'mail'}

    # -------------------------------------------------------------------------
    @classmethod
    def record_format(cls, record):
        """Check and possibly correct a record before inserting it in the
        database.

        :param dict record:
            Dictionary of values to check.
        :rtype: ``None`` or :class:`pyramid.i18n.TranslationString`
        :return:
            ``None`` or error message.
        """
        for k in [i for i in record if record[i] is None]:
            del record[k]

        password = record.get('root_password')
        if password and not record.get('root_dn'):
            del record['root_password']
        elif password and len(password) < 44:
            record['root_password'] = encrypt(password, 'ldap')

        if not record.get('base'):
            return _('No base for search.')

        if not record.get('user_dn'):
            return _('No DN for user.')
        if '_UID_' not in record['user_dn']:
            return _('_UID_ not found in user DN.')
        return None

    # -------------------------------------------------------------------------
    def db2xml(self):
        """Serialize LDAP settings to a XML representation.

        :rtype: lxml.etree.Element
        """
        ldap_elt = etree.Element('{{{0}}}ldap'.format(
            RELAXNG_CIOLDAP['namespace']))
        etree.SubElement(ldap_elt, 'host').text = self.host
        etree.SubElement(ldap_elt, 'port').text = str(self.port)
        if self.ssl:
            etree.SubElement(ldap_elt, 'ssl').text = 'true'
        etree.SubElement(ldap_elt, 'check-interval').text = \
            str(self.check_interval)
        if self.root_dn:
            etree.SubElement(ldap_elt, 'dn').text = self.root_dn
            etree.SubElement(ldap_elt, 'password').text = self.root_password
        etree.SubElement(ldap_elt, 'base').text = self.base

        user_elt = etree.SubElement(ldap_elt, 'user')
        etree.SubElement(user_elt, 'dn').text = self.user_dn
        etree.SubElement(user_elt, 'filter').text = self.user_filter
        if self.field_first_name:
            etree.SubElement(user_elt, 'first-name').text = \
                self.field_first_name
        etree.SubElement(user_elt, 'last-name').text = self.field_last_name
        etree.SubElement(user_elt, 'email').text = self.field_email
        if self.user_profiles:
            elt = etree.SubElement(user_elt, 'profiles')
            for dbprofile in self.user_profiles:
                etree.SubElement(elt, 'profile').text = dbprofile.profile_id

        return ldap_elt

    # -------------------------------------------------------------------------
    @classmethod
    def attachments2directory(cls, attachments, directory):
        """Copy from attachments directory the file corresponding to the
        ldap.

        :param str attachments:
            Absolute path to the attachements directory.
        :param str directory:
            The backup directory.
        """

    # -------------------------------------------------------------------------
    def sheet4view(self, request, form, profile_labels):
        """Generate the sheet content.

        :type  request: pyramid.request.Request
        :param request:
            Current request.
        :type  form: .lib.form.Form
        :param form:
            Current form object.
        :parem dict profile_labels:
            Label in user language of all profile.
        :rtype: webhelpers2.html.literal
        """
        translate = request.localizer.translate
        html = form.grid_item(translate(_('Host:')), self.host, clear=True)
        html += form.grid_item(translate(_('Port:')), self.port, clear=True)
        html += form.grid_item(
            translate(_('SSL:')), self.ssl and translate(_('yes')), clear=True)
        html += form.grid_item(
            translate(_('Check interval:')),
            deltatime_label(
                seconds=self.check_interval or 0, lang=request.locale_name),
            clear=True)
        html += form.grid_item(
            translate(_('Root DN:')), self.root_dn, clear=True)
        html += form.grid_item(translate(_('Base:')), self.base, clear=True)

        html += HTML.h3(translate(_('User')))
        html += form.grid_item(
            translate(_('User DN:')), self.user_dn, clear=True)
        html += form.grid_item(
            translate(_('User filter:')), self.user_filter, clear=True)
        html += form.grid_item(
            translate(_('First name field:')), self.field_first_name,
            clear=True)
        html += form.grid_item(
            translate(_('Last name field:')), self.field_last_name, clear=True)
        html += form.grid_item(
            translate(_('Email field:')), self.field_email, clear=True)
        if self.user_profiles:
            labels = [
                profile_labels[k.profile_id]
                for k in self.user_profiles if k.profile_id in profile_labels]
            html += form.grid_item(
                translate(_('Attributed Profiles:')), '<br/>'.join(labels),
                clear=True)

        return html

    # -------------------------------------------------------------------------
    @classmethod
    def settings_schema(cls, profiles, dbldap=None):
        """Return a Colander schema to edit LDAP.

        :param dict profiles:
            A dictionary such as ``{profile_id: label,...}``.
        :type  dbldap: DBLdap
        :param dbldap: (optional)
            Current CIOLdap SqlAlchemy object.
        :rtype: tuple
        :return:
            A tuple such as ``(schema, defaults)``.
        """
        # Schema
        schema = colander.SchemaNode(colander.Mapping())
        schema.add(colander.SchemaNode(
            colander.String(), name='host',
            validator=colander.Length(max=128), missing='localhost'))
        schema.add(colander.SchemaNode(
            colander.Integer(), name='port', missing=389))
        schema.add(colander.SchemaNode(
            colander.Boolean(), name='ssl', missing=False))
        schema.add(colander.SchemaNode(
            colander.Integer(), name='check_interval',
            missing=LDAP_CHECK_INTERVAL))
        schema.add(colander.SchemaNode(
            colander.String(), name='root_dn',
            validator=colander.Length(max=128), missing=''))
        schema.add(colander.SchemaNode(
            colander.String(), name='root_password1',
            validator=colander.Length(max=43), missing=None))
        schema.add(colander.SchemaNode(
            colander.String(), name='base',
            validator=colander.Length(max=128)))

        schema.add(colander.SchemaNode(
            colander.String(), name='user_dn',
            validator=colander.Length(max=128)))
        schema.add(colander.SchemaNode(
            colander.String(), name='user_filter',
            validator=colander.Length(max=128), missing=LDAP_USER_FILTER))
        schema.add(colander.SchemaNode(
            colander.String(), name='field_first_name',
            validator=colander.Length(max=32), missing=LDAP_FIELD_FIRSTNAME))
        schema.add(colander.SchemaNode(
            colander.String(), name='field_last_name',
            validator=colander.Length(max=32), missing=LDAP_FIELD_LASTNAME))
        schema.add(colander.SchemaNode(
            colander.String(), name='field_email',
            validator=colander.Length(max=32), missing=LDAP_FIELD_EMAIL))

        # Profiles
        for profile_id in profiles:
            schema.add(colander.SchemaNode(
                colander.Boolean(), name='pfl:{0}'.format(profile_id),
                missing=False))

        # Defaults
        defaults = {}
        if dbldap is None:
            defaults = {
                'host': 'localhost', 'port': 389,
                'check_interval': LDAP_CHECK_INTERVAL,
                'user_filter': LDAP_USER_FILTER,
                'field_first_name': LDAP_FIELD_FIRSTNAME,
                'field_last_name': LDAP_FIELD_LASTNAME,
                'field_email': LDAP_FIELD_EMAIL}
        else:
            defaults = {}
            for dbprofile in dbldap.user_profiles:
                defaults['pfl:{0}'.format(dbprofile.profile_id)] = True

        return schema, defaults

    # -------------------------------------------------------------------------
    @classmethod
    def sheet4edit(cls, request, form, profiles):
        """Generate the tab content for edition.

        :type  request: pyramid.request.Request
        :param request:
            Current request.
        :type  form: .lib.form.Form
        :param form:
            Current form object.
        :param dict profiles:
            A dictionary such as ``{profile_id: label,...}``.
        :rtype: webhelpers2.html.literal
        """
        translate = request.localizer.translate
        html = form.grid_text(
            'host', translate(_('Host:')), maxlength=128, required=True,
            clear=True)
        html += form.grid_text(
            'port', translate(_('Port:')), maxlength=8, required=True,
            clear=True)
        html += form.grid_custom_checkbox(
            'ssl', translate(_('SSL:')), clear=True)
        html += form.grid_text(
            'check_interval', translate(_('Check interval:')), maxlength=8,
            required=True, hint=translate(_('In seconds.')), clear=True)
        html += form.grid_text(
            'root_dn', translate(_('Root DN:')), maxlength=128, clear=True)
        html += form.grid_password(
            'root_password1', translate(_('Root password:')), maxlength=64,
            clear=True)
        html += form.grid_text(
            'base', translate(_('Base:')), maxlength=128, required=True,
            clear=True)

        html += HTML.h3(translate(_('User')))
        html += form.grid_text(
            'user_dn', translate(_('User DN:')), maxlength=128, required=True,
            clear=True)
        html += form.grid_text(
            'user_filter', translate(_('User filter:')), maxlength=128,
            required=True, hint=translate(_('Use _UID_ in place of user ID.')),
            clear=True)
        html += form.grid_text(
            'field_first_name', translate(_('First name field:')),
            maxlength=32, clear=True)
        html += form.grid_text(
            'field_last_name', translate(_('Last name field:')),
            maxlength=32, required=True, clear=True)
        html += form.grid_text(
            'field_email', translate(_('Email field:')), maxlength=32,
            clear=True)
        if profiles:
            html += literal(
                '<div class="cioFormItem">'
                '<label><strong>{0}</strong></label><div><ul>'.format(
                    translate(_('Attributed Profiles:'))))
            for profile_id in sorted(profiles):
                html += literal('<li>{0} {1}</li>'.format(
                    form.custom_checkbox(
                        'pfl:{0}'.format(profile_id)),
                    profiles[profile_id]))
            html += literal('</ul></div><div class="cioClear"></div></div>')

        return html


# =============================================================================
class DBLdapProfile(DBDeclarativeClass):
    """Class to link LDAP with its profiles (one-to-many)."""
    # pylint: disable = too-few-public-methods

    __tablename__ = 'ldap_profiles'
    __table_args__ = {'mysql_engine': 'InnoDB'}

    ldap_id = Column(
        Integer, ForeignKey('ldap.ldap_id', ondelete='CASCADE'),
        primary_key=True)
    profile_id = Column(
        String(ID_LEN), ForeignKey('profiles.profile_id', ondelete='CASCADE'),
        primary_key=True)



"""Class to manage user menus."""

from webhelpers2.html import literal

from ..includes.themes import theme_static_prefix
from .utils import make_id
from .breadcrumbs import DEFAULT_ROOT_CHUNKS


# =============================================================================
class Menu(object):
    """User menu base class.

    :type  request: pyramid.request.Request
    :param request:
        Current request.
    :param str uid:
        Unique ID used to store the menu in the session.
    :param list framework:
        Full version of the menu.

    The framework of the menu is a list of pieces. Each piece has the
    following structure:

        piece = (icon, label, permission, route, (subentry, ...))

    ``route`` can be a name of route or a tuple such as
    ``(route_name, params_dict)``.

    In the user session, under ``uid`` key, pieces are filtered according to
    user permissions and converted into menu entries. In an entry:

    * ``label`` is translated
    * ``permission`` is replaced by the final route.
    """

    # -------------------------------------------------------------------------
    def __init__(self, request, uid, framework):
        """Constructor method."""
        self._request = request
        self.uid = uid
        self._framework = framework

    # -------------------------------------------------------------------------
    def is_empty(self):
        """Return ``True`` if current menu is empty.

         :rtype: bool
        """
        self._construct()
        return not bool(self._request.session[self.uid])

    # -------------------------------------------------------------------------
    def xhtml(self, **kwargs):
        """Return an <ul> structure with current entry highlighted.

        :rtype:  webhelpers2.html._literal.literal

        Options are:

        - no_icon=True: suppress icons in entries
        - tooltip=True: add attribute 'title' on entries span
        """
        self._construct()

        crumb_trail = \
            (hasattr(self._request, 'breadcrumbs') and
             self._request.breadcrumbs.crumb_trail()) or \
            (self._request.matched_route is not None and
             [(self._request.current_route_path().partition('?')[0].split('/')[
                 1:DEFAULT_ROOT_CHUNKS + 1], DEFAULT_ROOT_CHUNKS)]) or []

        html = self._xhtml_entries(
            theme_static_prefix(self._request), 0,
            self._request.session[self.uid], crumb_trail, kwargs)

        return literal(
            '<ul id="cioMenu-{uid}" role="menu">{menu}</ul>'.format(
                uid=self.uid, menu=html)) if html else ''

    # -------------------------------------------------------------------------
    @classmethod
    def invalidate(cls, request, uid):
        """Invalidate the menu ``uid`` if exists.

        :type  request: pyramid.request.Request
        :param request:
            Current request.
        :param str uid:
            Unique ID of the menu to invalidate.
        """
        if uid in request.session:
            del request.session[uid]

    # -------------------------------------------------------------------------
    def _construct(self):
        """If the menu does not exist in session, construct it according to
        user permissions and put it in ``session[self.uid]``."""
        # Menu in session
        if self.uid in self._request.session:
            return

        # Construction
        menu = []
        for piece in self._framework:
            if self._has_permission(piece):
                entry = self._entry(piece)
                if entry is not None:
                    menu.append(entry)

        self._request.session[self.uid] = tuple(menu)

    # -------------------------------------------------------------------------
    def _xhtml_entries(self, theme, depth, entries, crumb_trail, options):
        """Return <li> tags with entries.

        :param str theme:
             URL prefix for the current theme.
        :param int depth:
            Depth of entries in menu.
        :param tuple entries:
            Tuple of entry tuples (See :meth:`_entry`)
        :param list crumb_trail:
            List of crumbs in the breadcrumb trail
            (See :meth:`~chrysalio.lib.breadcrumbs.Breadcrumbs.crumb_trail`).
        :param dict options:
            Dictionary of options to configure the menu.
        """
        html = ''
        for entry in entries:
            html += '<li class="cioDepth{depth}">'\
                '<span class="cioMenu-{class_}{current}"{title}>'.format(
                    class_=entry[0],
                    depth=depth,
                    current=' cioCurrent' if self._is_current(
                        crumb_trail, entry) else '',
                    title=' title="{0}"'.format(entry[2])
                    if options or options.get('tooltip') else '')
            # URL
            if entry[3]:
                html += '<a href="{url}"{blank}>'.format(
                    url=entry[3],
                    blank=entry[4] == 'blank' and ' target="_blank"' or '')
            # icon
            if entry[1] and (not options or not options.get('no_icon')):
                html += '<img src="{icon}" alt="{label}"/> '.format(
                        icon=entry[1].format(theme=theme), label=entry[2])
            html += '<span><span>{label}</span></span>'.format(label=entry[2])
            if entry[3]:  # URL
                html += '</a>'
            html += '</span>'
            if entry[5]:  # Subentries
                html += '<ul>{0}</ul>'.format(self._xhtml_entries(
                    theme, depth + 1, entry[5], crumb_trail, options))
            html += '</li>'

        return html

    # -------------------------------------------------------------------------
    @classmethod
    def _is_current(cls, crumb_trail, entry):
        """Check if the entry has to be highlighted.

        :param listt crumb_trail:
            A trail of crumbs to compare with menu entries.
            See :meth:`.lib.breadcrumbs.Breadcrumbs.crumb_trail`.`
        :param tuple entry:
            Entry to check.
        :rtype: bool
        """
        # Compare with the current path
        if not crumb_trail:
            return False
        crumb = crumb_trail[-1]
        entry_chunks = entry[3].split('/')[1:] if entry[3] else []
        if entry_chunks[:crumb[1]] == crumb[0]:
            return True
        for subentry in entry[5] or '':
            if cls._is_current(crumb_trail, subentry):
                return True

        # Compare with the paths of the breadcrumb trail
        for crumb in crumb_trail[1:-1]:
            if entry_chunks[:crumb[1]] == crumb[0]:
                return True
        return False

    # -------------------------------------------------------------------------
    def _entry(self, piece):
        """Construct a menu entry tuple and recursively the subentries.

        :param tuple piece:
            The piece of the menu framework.
        :rtype: tuple
        :return:
            A tuple such as
            ``(class, icon, label, route_path, route, (subentry, ...))``.
        """
        subentries = []
        if piece[4] is not None:
            for subpiece in piece[4]:
                if self._has_permission(subpiece):
                    subentry = self._entry(subpiece)
                    if subentry is not None:
                        subentries.append(subentry)

        route_path = None
        if piece[3]:
            route_path = (piece[3], {}) \
                if not isinstance(piece[3], (list, tuple)) else piece[3]
            route_path = self._request.route_path(
                route_path[0], **route_path[1])

        if route_path is not None or subentries:
            return (
                make_id(piece[1], mode='class', truncate=16).lower(),
                piece[0], self._request.localizer.translate(piece[1]),
                route_path, piece[3], subentries and tuple(subentries) or None)

        return None

    # -------------------------------------------------------------------------
    def _has_permission(self, piece):
        """Check if the user has the permission to use this piece of the menu
        framework.

        :param tuple piece:
            The piece of the menu framework to check.
        :rtype: bool
        """
        if piece[2] is None:
            return True
        return self._request.has_permission(piece[2]).boolval



# -*- coding: utf-8 -*-
# pylint: disable = import-outside-toplevel
"""Tests of ``views.settings`` class."""

from . import DBUnitTestCase, FunctionalTestCase


# =============================================================================
class UViewsSettingsSettingsView(DBUnitTestCase):
    """Unit test class for testing :class:`views.settings.SettingsView`."""

    # -------------------------------------------------------------------------
    def setUp(self):
        """Set up test application."""
        from ..models.dbsettings import DBSettings
        super(UViewsSettingsSettingsView, self).setUp()

        dbsession = self.request.dbsession
        dbsession.add(DBSettings(key='title', value='Test Chrysalio'))
        dbsession.add(DBSettings(key='email', value='admin@chrysal.io'))
        dbsession.add(DBSettings(key='password-min-length', value='8'))
        dbsession.add(DBSettings(key='language', value='en'))
        dbsession.add(DBSettings(key='page-size', value='80'))
        dbsession.add(DBSettings(key='download-max-size', value='10485760'))
        dbsession.add(DBSettings(key='clipboard-size', value='7'))

        self.configurator.add_route('home', '/')
        self.configurator.add_route('settings_view', '/settings/view')
        self.configurator.add_route('settings_edit', '/settings/edit')

        self.request.registry['settings']['email'] = 'admin@chrysal.io'

    # -------------------------------------------------------------------------
    def test_view(self):
        """[u:views.settings.SettingsView.view]"""
        from collections import namedtuple
        from ..views.settings import SettingsView
        from . import DummyPOST

        self.request.matched_route = namedtuple('Route', 'name')(
            name='settings_view')

        response = SettingsView(self.request).view()
        self.assertIn('form', response)
        self.assertEqual(
            self.request.breadcrumbs.current_title(), 'General Settings')

        self.request.is_xhr = True
        self.request.POST = DummyPOST()
        response = SettingsView(self.request).view()
        self.assertIsInstance(response, dict)
        self.assertFalse(response)

    # -------------------------------------------------------------------------
    def test_view_export(self):
        """[u:views.settings.SettingsView.view] export settings"""
        from pyramid.response import Response
        from ..views.settings import SettingsView

        self.request.POST = {'exp!.x': True}

        response = SettingsView(self.request).view()
        self.assertIsInstance(response, Response)
        self.assertEqual(response.content_type, 'application/xml')
        self.assertIn(b'<title>Test Chrysalio</title>', response.body)

    # -------------------------------------------------------------------------
    def test_view_import(self):
        """[u:views.settings.SettingsView.view] import settings"""
        from cgi import FieldStorage
        from . import TEST1_SET_XML, DummyPOST
        from ..views.settings import SettingsView
        from ..models.dbsettings import DBSettings

        dbsession = self.request.dbsession
        dbsetting = dbsession.query(DBSettings).filter_by(
            key='page-size').first()
        self.assertIsNotNone(dbsetting)
        self.assertEqual(dbsetting.value, '80')

        self.request.POST = DummyPOST({'imp!.x': True})
        with open(TEST1_SET_XML, 'r') as hdl:
            input_file = FieldStorage()
            input_file.file = hdl
            input_file.filename = TEST1_SET_XML
            self.request.POST.multikeys = {'file': (input_file,)}
            response = SettingsView(self.request).view()
        self.assertIn('dbsettings', response)
        dbsetting = dbsession.query(DBSettings).filter_by(
            key='page-size').first()
        self.assertIsNotNone(dbsetting)
        self.assertEqual(dbsetting.value, '40')

    # -------------------------------------------------------------------------
    def test_edit(self):
        """[u:views.settings.SettingsView.edit]"""
        from collections import namedtuple
        from ..views.settings import SettingsView

        self.request.matched_route = namedtuple('Route', 'name')(
            name='settings_edit')

        response = SettingsView(self.request).edit()
        self.assertIn('form', response)
        self.assertIn('dbsettings', response)
        self.assertEqual(
            self.request.breadcrumbs.current_title(),
            'General Settings Edition')

    # -------------------------------------------------------------------------
    def test_edit_save(self):
        """[u:views.settings.SettingsView.edit] save"""
        from pyramid.httpexceptions import HTTPFound
        from ..views.settings import SettingsView

        self.request.POST = {
            'title': 'Test Chrysalio', 'email': 'admin@chrysal.io',
            'language': 'en', 'password-min-length': 8, 'remember-me': 5184000,
            'download-max-size': 10485760, 'clipboard-size': 7, 'sav!.x': True}

        SettingsView(self.request).edit()
        errors = self.request.session.pop_flash('alert')
        self.assertEqual(len(errors), 1)
        self.assertIn('Correct', errors[0])

        self.request.POST['page-size'] = 40
        self.assertIsInstance(SettingsView(self.request).edit(), HTTPFound)


# =============================================================================
class FViewsSettingsSettingsView(FunctionalTestCase):
    """Functional test class for testing
    :class:`views.settings.SettingsView`."""

    # -------------------------------------------------------------------------
    def test_view(self):
        """[f:views.settings.SettingsView.view]"""
        self.login('test1')
        self.testapp.get('/settings/view', status=403)

        self.login('admin')
        response = self.testapp.get('/settings/view', status=200)
        self.assertIn(b'Edit', response.body)

    # -------------------------------------------------------------------------
    def test_edit(self):
        """[f:views.settings.SettingsView.edit]"""
        self.login('test1')
        self.testapp.get('/settings/edit', status=403)

        self.login('admin')
        response = self.testapp.get('/settings/edit', status=200)
        self.assertIn(b'Save', response.body)



"""An exemple of a module which generates conflict with SkeletonModule."""

from ...modules import Module


# =============================================================================
def includeme(configurator):
    """Function to register the module.

    :type  configurator: :pyramid.config.Configurator
    :param configurator:
        Object used to do configuration declaration within the application.
    """
    # Registration
    Module.register(configurator, ConflictModule)


# =============================================================================
class ConflictModule(Module):
    """Class for Chrysalio module."""

    name = 'Conflict'
    implements = ('skeleton',)



# pylint: disable = import-outside-toplevel
"""Tests of ``lib.menu`` class."""

from . import ConfiguratorTestCase

# Menu entry = (icon, label, permission, route, (subentry, ...))
MENU_HOME = ('{theme}/images/menu_home.png', 'Home', None, 'home', None)
MENU_ADMIN = (
    '{theme}/images/menu_admin.png', 'Administration', None, None,
    ((None, 'Profile Management', 'profile-create', 'profile_index', None),
     (None, 'User Management', None, 'user_index',
      ((None, 'Administrator', None, ('user_view', {'user_id': 1}), None),
       (None, 'User 1', None, ('user_view', {'user_id': 3}), None),
       (None, 'Nobody', None, None, None)))))


# =============================================================================
class ULibMenuMenu(ConfiguratorTestCase):
    """Unit test class for :class:`lib.menu.Menu`."""

    # -------------------------------------------------------------------------
    def setUp(self):
        """Set up test application."""
        from collections import namedtuple
        from pyramid.testing import DummyRequest
        from ..lib.breadcrumbs import Breadcrumbs

        super(ULibMenuMenu, self).setUp()
        self.configurator.add_route('home', '/')
        self.configurator.add_route('user_index', '/user/index')
        self.configurator.add_route('user_view', '/user/view/{user_id}')
        self.configurator.add_route('profile_index', '/profile/index')
        self.configurator.include('..security')

        self.request = DummyRequest(
            matched_route=namedtuple('Route', 'name')(name='home'))
        self.request.breadcrumbs = Breadcrumbs(self.request)
        self.request.registry['themes'] = {'': {'path': None, 'name': ''}}
        self.request.registry['settings'] = {'theme': ''}

        self.request.breadcrumbs('Home', forced_route=('home', {}))
        self.request.breadcrumbs(
            'User Management', forced_route=('user_index', {}))

    # -------------------------------------------------------------------------
    def test_is_empty(self):
        """[u:lib.menu.Menu.is_empty]"""
        from ..lib.menu import Menu

        self.assertTrue(Menu(self.request, 'foo', '').is_empty())
        self.assertFalse(
            Menu(self.request, 'menu', [MENU_HOME, MENU_ADMIN]).is_empty())

    # -------------------------------------------------------------------------
    def test_xhtml(self):
        """[u:lib.menu.Menu.xhtml]"""
        from ..lib.menu import Menu

        menu = Menu(self.request, 'menu', [MENU_HOME, MENU_ADMIN])
        html = menu.xhtml()
        self.assertIn('cioCurrent', html)
        self.assertNotIn('/profile/index', html)
        self.assertIn('/user/index', html)

        # Cache
        html = menu.xhtml()
        self.assertIn('/user/index', html)

    # -------------------------------------------------------------------------
    def test_xhtml_whitout_breadcrumbs(self):
        """[u:lib.menu.Menu.xhtml] without breadcrumbs"""
        from ..lib.menu import Menu

        del self.request.session['breadcrumbs']
        self.request.matched_route = None
        html = Menu(self.request, 'menu', [MENU_HOME, MENU_ADMIN]).xhtml()
        self.assertNotIn('cioCurrent', html)

    # -------------------------------------------------------------------------
    def test_invalidate(self):
        """[u:lib.menu.Menu.invalidate]"""
        from ..lib.menu import Menu

        self.request.session['menu'] = ()
        Menu.invalidate(self.request, 'menu')
        self.assertNotIn('menu', self.request.session)



"""Class to manage LDAP as an authority for authentication."""

from datetime import datetime

from ldap3 import Server, Connection
from ldap3.core.exceptions import LDAPBindError, LDAPSocketOpenError
from ldap3.core.exceptions import LDAPObjectClassError
from ldap3.core.exceptions import LDAPPasswordIsMandatoryError

from ....lib.i18n import _
from ....lib.utils import decrypt
from ....lib.config import settings_get_list
from ....lib.log import log_error
from ....models.dbprofile import DBProfile
from ....models.dbuser import DBUserProfile
from ..models.dbldap import DBLdap


# =============================================================================
class LDAP(object):
    """Authority class to manage LDAP authentication.

    It reads its parameters in its SQL table. The default values are:

    * host: localhost
    * port: 389
    * ssl: false
    * check_interval: 0 (no cache)
    * user_filter: (&(objectclass=inetOrgPerson)(uid=_UID_))
    * field_firstname: givenName
    * field_lastname: sn
    * field_email: mail
    """

    # -------------------------------------------------------------------------
    def __init__(self):
        """Constructor method."""
        self._settings = {
            'host': 'localhost',
            'port': 389,
            'ssl': False,
            'check_interval': 0,
            'root_dn': None,
            'root_password': None,
            'base': None,
            'user_dn': None,
            'user_filter': '(&(objectclass=inetOrgPerson)(uid={uid}))',
            'field_first_name': 'givenName',
            'field_last_name': 'sn',
            'field_email': 'mail',
            'user_profiles': []}

    # -------------------------------------------------------------------------
    def get(self, request, login, password, dbuser_class):
        """Get user from LDAP server.

        :type  request: pyramid.request.Request
        :param request:
            Current request.
        :param str login:
            Login of the user to authenticate.
        :param str password:
            Clear password.
        :type  dbuser_class: .models.dbuser.DBUser
        :param dbuser_class:
            The SQL class to create a user.
        :rtype: tuple
        :return:
            A tuple like ``(dbuser, error)``. It can be ``(None, None)``
        """
        if not password:
            return None, _('Password is mandatory.')
        self._load_configuration(request)
        if self._settings['base'] is None:
            return None, _('LDAP configuration is missing.')

        # Create user
        record, error = self._user_record(request, login, password)
        if error:
            return None, error
        now = datetime.now()
        language = request.accept_language.lookup(
            settings_get_list(
                request.registry.settings, 'languages', ['en']),
            default_tag=request.registry['settings']['language'])
        record.update({
            'login': login, 'password': password,
            'language': language, 'last_login': now,
            'account_creation': now, 'authority': 'ldap',
            'authority_check': now})
        error = dbuser_class.record_format(record)
        if error:  # pragma: nocover
            return None, error
        dbuser = dbuser_class(**record)
        request.dbsession.add(dbuser)
        request.dbsession.flush()

        # Add profiles
        if self._settings['user_profiles']:
            available = [k[0] for k in request.dbsession.query(
                DBProfile.profile_id).all()]
            for profile_id in self._settings['user_profiles']:
                if profile_id not in available:
                    log_error(
                        request, 'unknown profile {0}'.format(profile_id))
                else:
                    request.dbsession.add(DBUserProfile(
                        user_id=dbuser.user_id, profile_id=profile_id))

        return dbuser, None

    # -------------------------------------------------------------------------
    def check(self, request, login, password, dbuser):
        """Check user authorization according to LDAP server.

        :type  request: pyramid.request.Request
        :param request:
            Current request.
        :param str login:
            Login of the user to authenticate.
        :param str password:
            Clear password.
        :type  dbuser: .models.dbuser.DBUser
        :param dbuser:
            The SQL object of the user to check.
        :rtype: :class:`pyramid.i18n.TranslationString` or ``None``
        :return:
            Error message or ``None``.
        """
        self._load_configuration(request)
        if self._settings['base'] is None:
            return _('LDAP configuration is missing.')

        if dbuser.authority_check and \
           (datetime.now() - dbuser.authority_check).total_seconds() \
           < self._settings['check_interval']:
            return None

        # Information update
        record, error = self._user_record(request, login, password)
        if error:
            return error

        # Update database
        modified = False
        for field in record:
            if getattr(dbuser, field) != record[field]:
                setattr(dbuser, field, record[field])
                modified = True
        if modified:
            dbuser.account_update = datetime.now()
        dbuser.authority_check = datetime.now()
        dbuser.last_login = datetime.now()
        return None

    # -------------------------------------------------------------------------
    def reset_configuration(self):
        """Reset the configuration."""
        self._settings['base'] = None

    # -------------------------------------------------------------------------
    def _user_record(self, request, login, password):
        """Check user authorization according to LDAP server and retrieve
        information about the user.

        :type  request: pyramid.request.Request
        :param request:
            Current request.
        :param str login:
            User login.
        :param str password:
            Clear password.
        :rtype: tuple
        :return:
            A tuple such as ``(record, error)`` where ``record`` is a
            dictionary containing user information.
        """
        # Connection to LDAP server
        connection = self._ldap_connection(request, login, password)
        if connection is None:
            return None, _('ID or password is incorrect.')

        # Retrieve information from LDAP server
        fields = ('last_name', 'first_name', 'email')
        fields = [k for k in fields if self._settings['field_{0}'.format(k)]]
        try:
            connection.search(
                self._settings['base'],
                self._settings['user_filter'].format(uid=login),
                attributes=[
                    self._settings['field_{0}'.format(k)] for k in fields])
        except LDAPObjectClassError:
            return None, _('LDAP: incorrect filter "${f}".',
                           {'f': self._settings['user_filter']})
        if len(connection.entries) != 1:
            return None, _('LDAP: unable to retrieve user information.')

        attributes = connection.entries[0].entry_raw_attributes
        record = {
            k: attributes[
                self._settings['field_{0}'.format(k)]][0].decode('utf8')
            for k in fields
            if attributes[self._settings['field_{0}'.format(k)]]}

        return record, None

    # -------------------------------------------------------------------------
    def _ldap_connection(self, request, login, password):
        """Establish a connection with the LDAP server.

        :type  request: pyramid.request.Request
        :param request:
            Current request.
        :param str login:
            User login.
        :param str password:
            Clear password.
        :rtype: ldap3.Connection
        """
        server = Server(
            self._settings['host'], self._settings['port'],
            use_ssl=self._settings['ssl'])
        try:
            if password is not None:
                connection = Connection(
                    server, self._settings['user_dn'].format(login),
                    password, auto_bind=True)
            elif self._settings['root_dn']:
                connection = Connection(
                    server, self._settings['root_dn'],
                    decrypt(self._settings['root_password'], 'ldap'),
                    auto_bind=True)
            else:
                connection = Connection(server, auto_bind=True)
        except (LDAPBindError, LDAPSocketOpenError,
                LDAPPasswordIsMandatoryError) as error:
            log_error(request, error)
            return None
        return connection

    # -------------------------------------------------------------------------
    def _load_configuration(self, request):
        """Possibly load configuration from database.

        :type  request: pyramid.request.Request
        :param request:
            Current request.
        """
        if self._settings['base'] is not None:
            return

        dbldap = request.dbsession.query(DBLdap).first()
        if dbldap is None:
            return

        self._settings = {k: getattr(dbldap, k) for k in self._settings}
        self._settings['port'] = int(self._settings['port'])
        self._settings['user_dn'] = self._settings['user_dn'].replace(
            '_UID_', '{0}')
        self._settings['user_filter'] = self._settings['user_filter'].replace(
            '_UID_', '{uid}')
        # pylint: disable = no-member
        self._settings['user_profiles'] = [
            k.profile_id for k in dbldap.user_profiles]



# pylint: disable = import-outside-toplevel
"""Tests of ``lib.tabet`` class."""

from unittest import TestCase


# =============================================================================
class ULibTabsetTabset(TestCase):
    """Unit test class for :class:`lib.tabet.Tabset` and
    :class:`lib.tabet.Tab`."""

    # -------------------------------------------------------------------------
    @classmethod
    def make_one(cls):
        """Create an instance of Tabset.

        rtype: Tabset
        """
        from pyramid.testing import DummyRequest
        from ..lib.tabset import Tabset

        return Tabset(
            DummyRequest(), 'tabUser', ('Information', 'Profiles'))

    # -------------------------------------------------------------------------
    def test_begin(self):
        """[u:lib.tabset.Tabset.begin]"""
        from webhelpers2.html import literal

        html = self.make_one().begin()
        self.assertIsInstance(html, literal)
        self.assertIn('id="tabUser"', html)
        self.assertIn('id="tabUser1"', html)
        self.assertIn('href="#tabUser1_"', html)

    # -------------------------------------------------------------------------
    def test_end(self):
        """[u:lib.tabset.Tabset.end]"""
        from webhelpers2.html import literal

        html = self.make_one().end()
        self.assertIsInstance(html, literal)
        self.assertEqual(html, '</div>\n')

    # -------------------------------------------------------------------------
    def test_tab_begin(self):
        """[u:lib.tabset.Tab.tab]"""
        from webhelpers2.html import literal
        from ..lib.tabset import Tab

        tab = self.make_one().tabs[0]
        self.assertIsInstance(tab, Tab)
        html = tab.begin()
        self.assertIsInstance(html, literal)
        self.assertIn('id="tabUser0_"', html)

    # -------------------------------------------------------------------------
    def test_tab_end(self):
        """[u:lib.tabset.Tab.end]"""
        from webhelpers2.html import literal

        tab = self.make_one().tabs[1]
        html = tab.end()
        self.assertIsInstance(html, literal)
        self.assertEqual(html, '</fieldset>\n')



"""CIOSkeleton menu."""

from ...lib.i18n import _


# Menu entry = (icon, label, permission, route, (subentry, ...))
MENU_CIOSKELETON = (
    '{theme}/cioskeleton/images/menu_cioskeleton.png', _('Skeleton'),
    'bone-view', 'bone_index', None)



"""CIOLDAP Relax NG."""

from os.path import dirname, join


RELAXNG_CIOLDAP = {
    'namespace': 'http://ns.chrysal.io/cioldap',
    'root': 'cioldap', 'version': '1.0',
    'file': join(dirname(__file__), 'RelaxNG', 'cioldap.rng')}



# pylint: disable = import-outside-toplevel
"""Tests of ``views.error`` functions."""

from pyramid.testing import DummyRequest
from pyramid.httpexceptions import HTTPFound

from . import ConfiguratorTestCase, FunctionalTestCase


# =============================================================================
class DummyException(object):
    """Dummy exception class."""
    status_int = None
    explanation = None
    comment = None

    # -------------------------------------------------------------------------
    def __init__(self, status_int, title=None):
        """Constructor method."""
        self.status_int = status_int
        self.title = title


# =============================================================================
class UViewsErrorErrorView(ConfiguratorTestCase):
    """Unit test class for testing :func:`views.error.error_view`."""

    # -------------------------------------------------------------------------
    def setUp(self):
        """Set up test application."""
        super(UViewsErrorErrorView, self).setUp()
        self.configurator.add_route('login', '/login')

    # -------------------------------------------------------------------------
    def test_not_authenticated(self):
        """[u:views.error.error_view] not authenticated."""
        from ..views.error import error_view

        request = DummyRequest()
        request.exception = DummyException(403)
        request.is_xhr = False
        response = error_view(request)
        self.assertIsInstance(response, HTTPFound)

    # -------------------------------------------------------------------------
    def test_is_xhr(self):
        """[u:views.error.error_view] is xhr."""
        from ..views.error import error_view

        request = DummyRequest()
        request.exception = DummyException(404)
        request.is_xhr = True
        response = error_view(request)
        self.assertIn('error', response)
        self.assertEqual(response['error'], 404)

    # -------------------------------------------------------------------------
    def test_forbidden(self):
        """[u:views.error.error_view] forbidden."""
        from ..views.error import error_view

        self.configurator.testing_securitypolicy('test1', permissive=True)
        request = DummyRequest()
        request.exception = DummyException(403)
        request.is_xhr = False
        response = error_view(request)
        self.assertIn('status', response)
        self.assertEqual(response['status'], 403)

    # -------------------------------------------------------------------------
    def test_not_found(self):
        """[u:views.error.error_view] not found."""
        from ..views.error import error_view

        self.configurator.testing_securitypolicy(permissive=True)
        request = DummyRequest()
        request.exception = DummyException(404)
        request.is_xhr = False
        response = error_view(request)
        self.assertIn('status', response)
        self.assertEqual(response['status'], 404)

    # -------------------------------------------------------------------------
    def test_bad_request(self):
        """[u:views.error.error_view] bad request."""
        from ..views.error import error_view

        self.configurator.testing_securitypolicy(permissive=True)
        request = DummyRequest()
        request.exception = DummyException(400, title='Bad request')
        request.is_xhr = False
        response = error_view(request)
        self.assertIn('status', response)
        self.assertEqual(response['status'], 400)


# =============================================================================
class FViewsErrorErrorView(FunctionalTestCase):
    """Functional test class for testing :func:`views.error.error_view`."""

    # -------------------------------------------------------------------------
    def test_not_authenticated(self):
        """[f:views.error.error_view] not authenticated."""
        response = self.testapp.get('/', status=302)
        self.assertEqual(response.location, 'http://localhost/login?next=%2F')

    # -------------------------------------------------------------------------
    def test_not_found(self):
        """[f:views.error.error_view] not found."""
        self.login('test1')
        self.testapp.get('/user/foo', status=404)

    # -------------------------------------------------------------------------
    def test_bad_request(self):
        """[f:views.error.error_view] bad request."""
        self.testapp.post(
            '/login', {'login': 'test1', 'password': 'test1pwd'}, status=400)

    # -------------------------------------------------------------------------
    def test_forbidden(self):
        """[f:views.error.error_view] forbidden."""
        self.login('test1')
        self.testapp.get('/user/index', status=403)
